# allgrid.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Functions, and AllGrid class for methods, used by all *ListGrid classes.

'All' means CQLListGrid, EngineListGrid, GameListGrid, and QueryListGrid.

See cql_gamelist_query.CQLGameListQuery class for methods identical in
these classes except EngineListGrid
"""
import tkinter


class AllGrid:
    """Provide methods with identical implementation for *ListGrid classes.

    The definition is stretched to include the on_configure_canvas and
    create_edit_dialogue methods where one class, different for each method,
    has additional behaviour.
    """

    def create_edit_dialog(
        self, instance, newobject, oldobject, showinitial, modal, title=""
    ):
        """Extend to do chess initialization."""
        for obj in (newobject, oldobject):
            if obj:
                self._set_grid_database(obj)
                obj.load_record((instance.key.pack(), instance.srvalue))
        super().create_edit_dialog(
            instance, newobject, oldobject, showinitial, modal, title=title
        )

    def fill_view(
        self,
        currentkey=None,
        down=True,
        topstart=True,
        exclude=True,
    ):
        """Delegate to superclass if database is open otherwise do nothing."""
        # Intend to put this in superclass but must treat the DataClient
        # objects being scrolled as a database to do this properly.  Do this
        # when these objects have been given a database interface used when
        # the database is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of
        # games from PGN files; which can take many hours.

        if self.get_database() is not None:
            super().fill_view(
                currentkey=currentkey,
                down=down,
                topstart=topstart,
                exclude=exclude,
            )

    def load_new_index(self):
        """Delegate to superclass if database is open otherwise do nothing."""
        # Intend to put this in superclass but must treat the DataClient
        # objects being scrolled as a database to do this properly.  Do this
        # when these objects have been given a database interface used when
        # the database is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of
        # games from PGN files; which can take many hours.

        if self.get_database() is not None:
            super().load_new_index()

    def load_new_partial_key(self, key):
        """Delegate to superclass if database is open otherwise do nothing."""
        # Intend to put this in superclass but must treat the DataClient
        # objects being scrolled as a database to do this properly.  Do this
        # when these objects have been given a database interface used when
        # the database is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of
        # games from PGN files; which can take many hours.

        if self.get_database() is not None:
            super().load_new_partial_key(key)

    def on_configure_canvas(self, event=None):
        """Delegate to superclass if database is open otherwise do nothing."""
        # Intend to put this in superclass but must treat the DataClient
        # objects being scrolled as a database to do this properly.  Do this
        # when these objects have been given a database interface used when
        # the database is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of
        # games from PGN files; which can take many hours.

        if self.get_database() is not None:
            super().on_configure_canvas(event=event)
            self._set_object_panel_item_properties()

    def on_data_change(self, instance):
        """Delegate to superclass if database is open otherwise do nothing."""
        # Intend to put this in superclass but must treat the DataClient
        # objects being scrolled as a database to do this properly.  Do this
        # when these objects have been given a database interface used when
        # the database is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of
        # games from PGN files; which can take many hours.

        if self.get_database() is not None:
            super().on_data_change(instance)
            self._fill_view_from_top_hack()

    def is_payload_available(self):
        """Return True if grid is connected to a database."""
        data_source = self.get_data_source()
        if data_source is None:
            return False
        if data_source.get_database() is None:
            # Avoid exception scrolling visible grid not connected to database.
            # Make still just be hack to cope with user interface activity
            # while importing data.
            self.clear_grid_keys()

            return False
        return True

    def _set_event_bindings_frame(self, bindings=(), switch=True):
        """Set bindings if switch is True or unset the bindings."""
        ste = self.try_event
        frm = self.frame
        for sequence, function in bindings:
            self.bind(
                frm,
                sequence[0],
                function=ste(function) if switch and function else "",
            )

    def _set_grid_database(self, object_):
        """Do nothing: *ListGrid class is expected to override if needed."""

    def _set_object_panel_item_properties(self):
        """Do nothing: *ListGrid class is expected to override if needed."""

    def _fill_view_from_top_hack(self):
        """Do nothing: *ListGrid class is expected to override if needed."""

    def _configure_frame(self):
        self.gcanvas.configure(takefocus=tkinter.FALSE)
        self.data.configure(takefocus=tkinter.FALSE)
        self.frame.configure(takefocus=tkinter.FALSE)
        self.hsbar.configure(takefocus=tkinter.FALSE)
        self.vsbar.configure(takefocus=tkinter.FALSE)



# chessdu.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Define User Interface for deferred update process."""

import os
import datetime
import tkinter
import tkinter.font
import tkinter.messagebox
import queue
import time
import multiprocessing
import multiprocessing.dummy

from solentware_misc.gui.logtextbase import LogTextBase

from solentware_bind.gui.bindings import Bindings

from pgn_read.core.parser import PGN

from solentware_base.core.constants import (
    BRECPPG,
    DEFAULT_RECORDS,
)

from ..core.pgn import GameUpdateEstimate
from .. import (
    ERROR_LOG,
    APPLICATION_NAME,
)
from ..core.filespec import (
    GAMES_FILE_DEF,
    BTOD_FACTOR,
)

# Time taken to parse a sample of a PGN file is measured.
# Number of games in file is estimated from number of bytes used in game scores
# compared with number of bytes in file.
# _DATABASE_UPDATE_FACTOR is set from measuring time taken to import a PGN file
# containing a large number of games to an empty database.  Large means over a
# million games.
# This factor is a big under-estimate when number of games is less than segment
# size.  In practice the difference will be noticed when importing less than
# 131072 games to a database which already has games: the new games will span
# up to three segments when imported.
# Available memory has a big impact because it determines how long the runs of
# sequential updates to indexes can be.  The factor 5 is appropriate when at
# least 1.5Gb is available.  A default build of FreeBSD 10.1 on a PC with 2Gb
# installed passes this test; but Microsoft Windows XP, and later presumably,
# needs more memory to do so.  A default build of OpenBSD 5.9 restricts user
# processes to 0.5Gb.  The situation is not known for OS X or any Linux
# distribution.
# Factor changed from 3 to 5 when CQL5.1 syntax introduced to implement partial
# position searches, due to extra index updates.
_DATABASE_UPDATE_FACTOR = 5


class _Reporter:
    """Helper class to keep 'LogText' API for adding text to log.

    Not used in dptcompatdu module but is used in chessdptdu module.

    """

    def __init__(self, append_text, append_text_only):
        """Note the timestamp plus text, and text only, append methods."""
        self.append_text = append_text
        self.append_text_only = append_text_only


class IncreaseDataProcess:
    """Define a process to do an increase data size (table B) process."""

    def __init__(self, database, report_queue, quit_event):
        """Provide queues for communication with GUI."""
        self.database = database
        self.report_queue = report_queue
        self.quit_event = quit_event
        self.process = multiprocessing.Process(
            target=self._increase_data_size,
            args=(),
        )
        self.stop_thread = None

    def _report_to_log(self, text):
        """Add text to report queue with timestamp."""
        day, hms = datetime.datetime.isoformat(
            datetime.datetime.today()
        ).split("T")
        hms = hms.split(".")[0]
        self.report_queue.put("".join((day, " ", hms, "  ", text, "\n")))

    def _report_to_log_text_only(self, text):
        """Add text to report queue without timestamp."""
        self.report_queue.put("".join(("                     ", text, "\n")))

    def _increase_data_size(self):
        """Increase data size."""
        self.stop_thread = multiprocessing.dummy.DummyProcess(
            target=self._wait_for_quit_event
        )
        self.stop_thread.start()
        self._increase()
        self.quit_event.set()

    def _wait_for_quit_event(self):
        """Wait for quit event."""
        self.quit_event.wait()

    def _increase(self):
        """Increase data size in DPT games file."""
        database = self.database
        files = (GAMES_FILE_DEF,)
        database.open_context_prepare_import(files=files)
        try:
            parameter = database.get_database_parameters(files=files)[
                GAMES_FILE_DEF
            ]
            bsize = parameter["BSIZE"]
            bused = max(0, parameter["BHIGHPG"])
            bfree = bsize - bused
            dsize = parameter["DSIZE"]
            dused = parameter["DPGSUSED"]
            dfree = dsize - dused
            table = database.table[GAMES_FILE_DEF]
            specification = database.specification[GAMES_FILE_DEF]
            default_records = specification[DEFAULT_RECORDS]
            btod_factor = specification[BTOD_FACTOR]
            brecppg = table.filedesc[BRECPPG]
            bdefault = database.get_pages_for_record_counts(
                (default_records, default_records)
            )[0]
            bfree_recs = bfree * brecppg
            dfree_recs = (dfree * brecppg) // btod_factor
            blow = bfree + min(bdefault, bfree)
            bhigh = bfree + max(bdefault, bfree)
            blowrecs = bfree_recs + min(default_records, bfree_recs)
            bhighrecs = bfree_recs + max(default_records, bfree_recs)
            if len(table.get_extents()) % 2 == 0:
                if not tkinter.messagebox.askokcancel(
                    title="Increase Data Size",
                    message="".join(
                        (
                            "At present it is better to do index ",
                            "increases first for this file, if any ",
                            "are needed.\n\nIt is estimated the ",
                            "current index size can cope with an ",
                            "estimated extra ",
                            str(dfree_recs),
                            " games.\n\nPlease confirm you wish to ",
                            "continue with data increase.",
                        )
                    ),
                ):
                    return
            if blow != bhigh:
                choice = tkinter.messagebox.askyesnocancel(
                    title="Increase Data Size",
                    message="".join(
                        (
                            "Please choose between a data size increase ",
                            "to cope with an estimated extra ",
                            str(blowrecs),
                            " or ",
                            str(bhighrecs),
                            " games.\n\nThe current data size can cope ",
                            "with an estimated extra ",
                            str(bfree_recs),
                            " games.\n\nDo you want to increase the data ",
                            "size for the smaller number of games?",
                        )
                    ),
                )
                if choice is True:  # Yes
                    bincrease = blow - bfree
                    bextrarecs = blowrecs
                elif choice is False:  # No
                    bincrease = bhigh - bfree
                    bextrarecs = bhighrecs
                else:  # Cancel assumed (choice is None).
                    return
            else:
                choice = tkinter.messagebox.askokcancel(
                    title="Increase Data Size",
                    message="".join(
                        (
                            "Please confirm a data size increase ",
                            "to cope with an estimated extra ",
                            str(blowrecs),
                            " games.\n\nThe current data size can cope ",
                            "with an estimated extra ",
                            str(bfree_recs),
                            " games.",
                        )
                    ),
                )
                if choice is True:  # Yes
                    bincrease = blow - bfree
                    bextrarecs = blowrecs
                else:  # Cancel assumed (choice is None).
                    return
            table.opencontext.Increase(bincrease, False)
            self._report_to_log_text_only("")
            self._report_to_log("Data size increased.")
            self._report_to_log_text_only(
                " ".join(
                    (
                        "Estimate of standard profile games which fit:",
                        str(bextrarecs),
                    )
                )
            )
        finally:
            database.close_database()


class IncreaseIndexProcess:
    """Define a process to do an increase index size (table D) process."""

    def __init__(self, database, report_queue, quit_event):
        """Provide queues for communication with GUI."""
        self.database = database
        self.report_queue = report_queue
        self.quit_event = quit_event
        self.process = multiprocessing.Process(
            target=self._increase_index_size,
            args=(),
        )
        self.stop_thread = None

    def _report_to_log(self, text):
        """Add text to report queue with timestamp."""
        day, hms = datetime.datetime.isoformat(
            datetime.datetime.today()
        ).split("T")
        hms = hms.split(".")[0]
        self.report_queue.put("".join((day, " ", hms, "  ", text, "\n")))

    def _report_to_log_text_only(self, text):
        """Add text to report queue without timestamp."""
        self.report_queue.put("".join(("                     ", text, "\n")))

    def _increase_index_size(self):
        """Increase index size."""
        self.stop_thread = multiprocessing.dummy.DummyProcess(
            target=self._wait_for_quit_event
        )
        self.stop_thread.start()
        self._increase()
        self.quit_event.set()

    def _wait_for_quit_event(self):
        """Wait for quit event."""
        self.quit_event.wait()

    def _increase(self):
        """Increase index size in DPT games file."""
        database = self.database
        files = (GAMES_FILE_DEF,)
        database.open_context_prepare_import(files=(GAMES_FILE_DEF,))
        try:
            parameter = database.get_database_parameters(files=files)[
                GAMES_FILE_DEF
            ]
            bsize = parameter["BSIZE"]
            bused = max(0, parameter["BHIGHPG"])
            bfree = bsize - bused
            dsize = parameter["DSIZE"]
            dused = parameter["DPGSUSED"]
            dfree = dsize - dused
            table = database.table[GAMES_FILE_DEF]
            specification = database.specification[GAMES_FILE_DEF]
            default_records = specification[DEFAULT_RECORDS]
            btod_factor = specification[BTOD_FACTOR]
            brecppg = table.filedesc[BRECPPG]
            ddefault = database.get_pages_for_record_counts(
                (default_records, default_records)
            )[1]
            bfree_recs = bfree * brecppg
            dfree_recs = (dfree * brecppg) // btod_factor
            dlow = dfree + min(ddefault, dfree)
            dhigh = dfree + max(ddefault, dfree)
            dlowrecs = dfree_recs + min(default_records, dfree_recs)
            dhighrecs = dfree_recs + max(default_records, dfree_recs)
            if len(table.get_extents()) % 2 != 0:
                if not tkinter.messagebox.askokcancel(
                    title="Increase Index Size",
                    message="".join(
                        (
                            "At present it is better to do data ",
                            "increases first for this file, if any ",
                            "are needed.\n\nIt is estimated the ",
                            "current data size can cope with an ",
                            "estimated extra ",
                            str(bfree_recs),
                            " games.\n\nPlease confirm you wish to ",
                            "continue with index increase.",
                        )
                    ),
                ):
                    return
            if dlow != dhigh:
                choice = tkinter.messagebox.askyesnocancel(
                    title="Increase Index Size",
                    message="".join(
                        (
                            "Please choose between an index size increase ",
                            "to cope with an estimated extra ",
                            str(dlowrecs),
                            " or ",
                            str(dhighrecs),
                            " games.\n\nThe current index size can cope ",
                            "with an estimated extra ",
                            str(dfree_recs),
                            " games.\n\nDo you want to increase the index ",
                            "size for the smaller number of games?",
                        )
                    ),
                )
                if choice is True:  # Yes
                    dincrease = dlow - dfree
                    dextrarecs = dlowrecs
                elif choice is False:  # No
                    dincrease = dhigh - dfree
                    dextrarecs = dhighrecs
                else:  # Cancel assumed (choice is None).
                    return
            else:
                choice = tkinter.messagebox.askokcancel(
                    title="Increase Index Size",
                    message="".join(
                        (
                            "Please confirm an index size increase ",
                            "to cope with an estimated extra ",
                            str(dlowrecs),
                            " games.\n\nThe current index size can cope ",
                            "with an estimated extra ",
                            str(dfree_recs),
                            " games.",
                        )
                    ),
                )
                if choice is True:  # Yes
                    dincrease = dlow - dfree
                    dextrarecs = dlowrecs
                else:  # Cancel assumed (choice is None).
                    return
            table.opencontext.Increase(dincrease, True)
            self._report_to_log_text_only("")
            self._report_to_log("Index size increased.")
            self._report_to_log_text_only(
                " ".join(
                    (
                        "Estimate of standard profile games which fit:",
                        str(dextrarecs),
                    )
                )
            )
        finally:
            database.close_database()


class DeferredUpdateProcess:
    """Define a process to do a deferred update task."""

    def __init__(
        self,
        database,
        method,
        report_queue,
        quit_event,
        increases,
        home_directory,
        pgnfiles,
    ):
        """Provide queues for communication with GUI."""
        self.database = database
        self.method = method
        self.report_queue = report_queue
        self.quit_event = quit_event
        self.increases = increases
        self.home_directory = home_directory
        self.pgnfiles = pgnfiles
        self.process = multiprocessing.Process(
            target=self._run_import,
            args=(),
        )

    def _report_to_log(self, text):
        """Add text to report queue with timestamp."""
        day, hms = datetime.datetime.isoformat(
            datetime.datetime.today()
        ).split("T")
        hms = hms.split(".")[0]
        self.report_queue.put("".join((day, " ", hms, "  ", text, "\n")))

    def _report_to_log_text_only(self, text):
        """Add text to report queue without timestamp."""
        self.report_queue.put("".join(("                     ", text, "\n")))

    def _run_import(self):
        """Invoke method to do the deferred update and display job status."""
        self.method(
            self.home_directory,
            self.pgnfiles,
            file=GAMES_FILE_DEF,
            reporter=_Reporter(
                self._report_to_log,
                self._report_to_log_text_only,
            ),
            quit_event=self.quit_event,
            increases=self.increases,
        )


class DeferredUpdateEstimateProcess:
    """Define a process to do a deferred update estimate task."""

    def __init__(
        self,
        database,
        sample,
        report_queue,
        quit_event,
        increases,
        pgnfiles,
    ):
        """Provide queues for communication with GUI."""
        self.database = database
        self.sample = sample
        self.report_queue = report_queue
        self.quit_event = quit_event
        self.increases = increases
        self.estimate_data = None
        self.pgnfiles = pgnfiles
        self.process = multiprocessing.Process(
            target=self._allow_import,
            args=(),
        )
        self.stop_thread = None

    def _report_to_log(self, text):
        """Add text to report queue with timestamp."""
        day, hms = datetime.datetime.isoformat(
            datetime.datetime.today()
        ).split("T")
        hms = hms.split(".")[0]
        self.report_queue.put("".join((day, " ", hms, "  ", text, "\n")))

    def _report_to_log_text_only(self, text):
        """Add text to report queue without timestamp."""
        self.report_queue.put("".join(("                     ", text, "\n")))

    def _wait_for_quit_event(self):
        """Wait for quit event."""
        self.quit_event.wait()

    def _allow_import(self):
        """Do checks for database engine and return True if import allowed."""
        self.stop_thread = multiprocessing.dummy.DummyProcess(
            target=self._wait_for_quit_event
        )
        self.stop_thread.start()
        # The close_database() in finally clause used to be the first statement
        # after runjob() definition in _run_import() method.  An exception was
        # raised using the sqlite3 module because run_input() is run in a
        # different thread from allow_input().  Earlier versions of chessdu did
        # not attempt to close the connection, hiding the problem.
        # The apsw module did not raise an exception, nor did modules providing
        # an interface to Berkeley DB or DPT.
        self.database.open_database()
        try:
            if not self._estimate_games_in_import():
                return None
            if self._allow_time():
                self.database.report_plans_for_estimate(
                    self._get_pgn_file_estimates(),
                    _Reporter(
                        self._report_to_log,
                        self._report_to_log_text_only,
                    ),
                    self.increases,
                )
                self.quit_event.set()
                return True
            self._report_to_log("Unable to estimate time to do import.")
            self._report_to_log_text_only("")
            return False
        finally:
            self.database.close_database()

    def _estimate_games_in_import(self):
        """Estimate import size from the first sample games in import files."""
        self.estimate_data = False
        text_file_size = sum(os.path.getsize(pp) for pp in self.pgnfiles)
        reader = PGN(game_class=GameUpdateEstimate)
        errorcount = 0
        totallen = 0
        totalerrorlen = 0
        totalgamelen = 0
        gamecount = 0
        positioncount = 0
        piecesquaremovecount = 0
        piecemovecount = 0
        estimate = False
        time_start = time.monotonic()
        for pgnfile in self.pgnfiles:
            if gamecount + errorcount >= self.sample:
                estimate = True
                break
            with open(pgnfile, "r", encoding="iso-8859-1") as source:
                for rcg in reader.read_games(source):
                    if self.quit_event.is_set():
                        self._report_to_log_text_only("")
                        self._report_to_log(
                            " ".join(
                                (
                                    "Estimating task stopped when",
                                    str(gamecount),
                                    "games and",
                                    str(errorcount),
                                    "games with errors found.",
                                )
                            )
                        )
                        return False
                    if gamecount + errorcount >= self.sample:
                        estimate = True
                        break
                    if len(rcg.pgn_text):
                        rawtokenlen = rcg.end_char - rcg.start_char
                    else:
                        rawtokenlen = 0
                    if rcg.state is not None:
                        errorcount += 1
                        totalerrorlen += rawtokenlen
                    else:
                        gamecount += 1
                        totalgamelen += rawtokenlen
                        positioncount += len(rcg.positionkeys)
                        piecesquaremovecount += len(rcg.piecesquaremovekeys)
                        piecemovecount += len(rcg.piecemovekeys)
                    totallen += rawtokenlen
        time_end = time.monotonic()
        if estimate:
            try:
                scale = float(text_file_size) // totallen
            except ZeroDivisionError:
                scale = 0
        else:
            scale = 1
        try:
            bytes_per_game = totalgamelen // gamecount
        except ZeroDivisionError:
            bytes_per_game = 0
        try:
            bytes_per_error = totalerrorlen // errorcount
        except ZeroDivisionError:
            bytes_per_error = 0
        try:
            positions_per_game = positioncount // gamecount
        except ZeroDivisionError:
            positions_per_game = 0
        try:
            pieces_per_game = piecesquaremovecount // gamecount
        except ZeroDivisionError:
            pieces_per_game = 0
        try:
            piecetypes_per_game = piecemovecount // gamecount
        except ZeroDivisionError:
            piecetypes_per_game = 0

        self.estimate_data = (
            int(gamecount * scale),
            bytes_per_game,
            positions_per_game,
            pieces_per_game,
            piecetypes_per_game,
            int(errorcount * scale),
            bytes_per_error,
            estimate,
            gamecount,
            errorcount,
            (time_end - time_start) / (gamecount + errorcount),
        )
        if estimate:
            self._report_to_log_text_only("")
            self._report_to_log(
                " ".join(("Estimated Games:", str(self.estimate_data[0])))
            )
            self._report_to_log_text_only(
                " ".join(("Estimated Errors:", str(self.estimate_data[5]))),
            )
            self._report_to_log_text_only(
                " ".join(("Sample Games:", str(gamecount)))
            )
            self._report_to_log_text_only(
                " ".join(("Sample Errors:", str(errorcount)))
            )
        else:
            self._report_to_log_text_only("")
            self._report_to_log(
                "The import is small so all games are counted."
            )
            self._report_to_log_text_only(
                " ".join(("Games in import:", str(gamecount)))
            )
            self._report_to_log_text_only(
                " ".join(("Errors in import:", str(errorcount))),
            )
        self._report_to_log_text_only(
            " ".join(("Bytes per game:", str(bytes_per_game)))
        )
        self._report_to_log_text_only(
            " ".join(("Bytes per error:", str(bytes_per_error))),
        )
        self._report_to_log_text_only(
            " ".join(("Positions per game:", str(positions_per_game))),
        )

        # positions_per_game == 0 if there are no valid games.
        ppp = "Not a number"
        for count, text in (
            (pieces_per_game, "Pieces per position:"),
            (piecetypes_per_game, "Piece types per position:"),
        ):
            if positions_per_game:
                ppp = str(count // positions_per_game)
            self._report_to_log_text_only(" ".join((text, ppp)))
        self._report_to_log_text_only("")

        # Check if import can proceed
        if gamecount + errorcount == 0:
            self._report_to_log(
                "No games, or games with errors, found in import."
            )
            return False
        if errorcount == 0:
            if estimate:
                self._report_to_log("No games with errors in sample.")
                self._report_to_log_text_only(
                    " ".join(
                        (
                            "It is estimated no games with",
                            "errors exist in import.",
                        )
                    ),
                )
                self._report_to_log_text_only(
                    "Any found in import will be indexed only as errors.",
                )
            else:
                self._report_to_log("No games with errors in import.")
            self._report_to_log_text_only("")
        elif estimate:
            self._report_to_log("Games with errors have been found in sample.")
            self._report_to_log_text_only(
                " ".join(
                    (
                        "The sample is the first",
                        str(self.sample),
                        "games in import.",
                    )
                ),
            )
            self._report_to_log_text_only(
                "All found in import will be indexed only as errors.",
            )
            self._report_to_log_text_only("")
        else:
            self._report_to_log("Games with errors have been found in sample.")
            self._report_to_log_text_only(
                "All found in import will be indexed only as errors.",
            )
            self._report_to_log_text_only("")
        return True

    def _allow_time(self):
        """Ask is deferred update to proceed if game count is estimated.

        The time taken will vary significantly depending on environment.

        """
        if not self.estimate_data:
            return False
        seconds = (
            (self.estimate_data[0] + self.estimate_data[5])
            * self.estimate_data[10]
            * _DATABASE_UPDATE_FACTOR
        )
        minutes, seconds = divmod(round(seconds), 60)
        hours, minutes = divmod(minutes, 60)
        days, hours = divmod(hours, 24)
        duration = []
        if days:
            duration.append(str(days))
            duration.append("days")
            if hours:
                duration.append(str(hours))
                duration.append("hours")
        elif hours:
            duration.append(str(hours))
            duration.append("hours")
            if minutes:
                duration.append(str(minutes))
                duration.append("minutes")
        elif minutes:
            duration.append(str(minutes))
            duration.append("minutes")
            if seconds:
                duration.append(str(seconds))
                duration.append("seconds")
        elif seconds > 1:
            duration.append(str(seconds))
            duration.append("seconds")
        else:
            duration.append(str(1))
            duration.append("second")
        self._report_to_log(
            "".join(
                (
                    "The estimate is the time taken to process the sample ",
                    "scaled up to the estimated number of games then ",
                    "multiplied by ",
                    str(_DATABASE_UPDATE_FACTOR),
                    ".  Expect the import to take longer if the database ",
                    "already contains games: the effect is worse for smaller "
                    "imports.  Progress reports are made if the import is ",
                    "large enough.",
                )
            )
        )
        self._report_to_log_text_only("")
        self._report_to_log(
            "".join(
                ("The import is expected to take ", " ".join(duration), ".")
            )
        )
        return True

    def _get_pgn_file_estimates(self):
        """Return the estimates of object counts for a PGN file."""
        return self.estimate_data


class ChessDeferredUpdate(Bindings):
    """Connect a chess database with User Interface for deferred update."""

    def __init__(
        self,
        deferred_update_method=None,
        database_class=None,
        home_directory=None,
        pgnfiles=None,
        sample=5000,
    ):
        """Create the database and ChessUI objects.

        deferred_update_method - the method to do the import
        database_class - access the database with an instance of this class
        sample - estimate size of import from first 'sample' games in PGN file

        """
        super().__init__()
        self.set_error_file_name(os.path.join(home_directory, ERROR_LOG))
        self.report_queue = multiprocessing.Queue()
        self.quit_event = multiprocessing.Event()
        self.increases = multiprocessing.Array("i", [0, 0, 0, 0])
        self.home_directory = home_directory
        self.pgnfiles = pgnfiles
        self.dumethod = deferred_update_method
        self.sample = sample
        self._import_done = False
        self._import_job = None
        self._task_name = "estimating"
        self.database = database_class(
            home_directory,
            allowcreate=True,
            deferupdatefiles={GAMES_FILE_DEF},
        )

        self.root = tkinter.Tk()
        self.root.wm_title(
            " - ".join(
                (
                    " ".join((APPLICATION_NAME, "Import")),
                    os.path.basename(home_directory),
                )
            )
        )
        frame = tkinter.Frame(master=self.root)
        frame.pack(side=tkinter.BOTTOM)
        # Not yet sure 'self.buttonframe' should become 'buttonframe'.
        self.buttonframe = tkinter.Frame(master=frame)
        self.buttonframe.pack(side=tkinter.BOTTOM)
        tkinter.Button(
            master=self.buttonframe,
            text="Dismiss Log",
            underline=0,
            command=self.try_command(
                self._dismiss_import_log,
                self.buttonframe,
            ),
        ).pack(side=tkinter.RIGHT, padx=12)
        tkinter.Button(
            master=self.buttonframe,
            text="Stop Process",
            underline=0,
            command=self.try_command(self._stop_task, self.buttonframe),
        ).pack(side=tkinter.RIGHT, padx=12)
        tkinter.Button(
            master=self.buttonframe,
            text="Import",
            underline=0,
            command=self.try_command(self._do_import, self.buttonframe),
        ).pack(side=tkinter.RIGHT, padx=12)
        if self._database_looks_like_dpt():
            tkinter.Button(
                master=self.buttonframe,
                text="Increase Index",
                underline=13,
                command=self.try_command(
                    self._increase_index, self.buttonframe
                ),
            ).pack(side=tkinter.RIGHT, padx=12)
            tkinter.Button(
                master=self.buttonframe,
                text="Increase Data",
                underline=9,
                command=self.try_command(
                    self._increase_data, self.buttonframe
                ),
            ).pack(side=tkinter.RIGHT, padx=12)

        self.report = LogTextBase(
            master=self.root,
            cnf=dict(wrap=tkinter.WORD, undo=tkinter.FALSE),
        )
        self.report.focus_set()
        if self._database_looks_like_dpt():
            self.bind(
                self.report,
                "<Alt-d>",
                function=self.try_event(self._increase_data),
            )
            self.bind(
                self.report,
                "<Alt-x>",
                function=self.try_event(self._increase_index),
            )
        self.bind(
            self.report,
            "<Alt-i>",
            function=self.try_event(self._do_import),
        )
        self.bind(
            self.report,
            "<Alt-d>",
            function=self.try_event(
                self._dismiss_import_log,
            ),
        )
        self.bind(
            self.report,
            "<Alt-s>",
            function=self.try_event(self._stop_task),
        )

        self.report.tag_configure(
            "margin",
            lmargin2=tkinter.font.nametofont(self.report.cget("font")).measure(
                "2010-05-23 10:20:57  "
            ),
        )
        self.tagstart = "1.0"
        self._report_to_log(
            "".join(("Importing to database ", home_directory, "."))
        )
        self._report_to_log_text_only(
            "All times quoted assume no other applications running.",
        )
        self._report_to_log_text_only("")
        self._report_to_log("Estimating number of games in import.")
        self.report.pack(
            side=tkinter.LEFT, fill=tkinter.BOTH, expand=tkinter.TRUE
        )
        self.root.iconify()
        self.root.update()
        self.root.deiconify()
        self._allow_job = False
        self.deferred_update = DeferredUpdateEstimateProcess(
            self.database,
            self.sample,
            self.report_queue,
            self.quit_event,
            self.increases,
            self.pgnfiles,
        )
        self.deferred_update.process.start()
        self.quit_thread = multiprocessing.dummy.DummyProcess(
            target=self._deferred_update_estimate_join
        )
        self.quit_thread.start()
        self._add_queued_reports_to_log()
        self.root.mainloop()

    def _database_looks_like_dpt(self):
        """Return True if database attribute signature looks like DPT.

        Check a few attriute names expected only in Database class in
        solentware_base.core._dpt module.

        """
        # This describes situation before changes to resolve problem, but
        # the return value remains relevant.
        # An alternative implementation of this difference calls a method
        # add_import_buttons() rather than add the buttons if the test
        # here returns True.  Two versions of add_import_buttons() are
        # defined in classes ..dpt.chessdptdu.ChessDatabase and
        # ..shared.dptcompatdu.DptCompatdu and the class hierarchy does
        # the test implemented here.  At present that implementation fails
        # because module pickling errors occur for the import action if
        # preceded by an increase action: but some solved problems in this
        # implementation hint at changes which might allow the alternative
        # implementation to succeed.  A practical benefit of the alternative
        # is losing the process startup overhead in the two (quite quick)
        # increase actions relevant only in DPT.
        return hasattr(self.database, "parms") and hasattr(
            self.database, "msgctl"
        )

    def _report_to_log(self, text):
        """Add text to report queue with timestamp."""
        day, hms = datetime.datetime.isoformat(
            datetime.datetime.today()
        ).split("T")
        hms = hms.split(".")[0]
        self.report_queue.put("".join((day, " ", hms, "  ", text, "\n")))

    def _report_to_log_text_only(self, text):
        """Add text to report queue without timestamp."""
        self.report_queue.put("".join(("                     ", text, "\n")))

    def _add_queued_reports_to_log(self):
        """Check report queue every 200ms and add reports to log."""
        # Items are put on queue infrequently relative to polling, so testing
        # the unreliable qsize() value is worthwhile because it will usually
        # be 0 thus avoiding the Empty exception.
        while self.report_queue.qsize():
            try:
                self.report.append_raw_text(self.report_queue.get_nowait())
            except queue.Empty:
                pass
        self.root.after(200, self._add_queued_reports_to_log)

    def _deferred_update_estimate_join(self):
        """join() deferred_update process then allow quit."""
        self.deferred_update.process.join()
        self._allow_job = True

    def _deferred_update_join(self):
        """join() deferred_update process then allow quit."""
        self.deferred_update.process.join()
        self._allow_job = True
        self._import_done = True

    def _increase_data(self, event=None):
        """Run Increase Data Size process."""
        del event
        if not self._allow_job:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Increase Data",
                message="".join(
                    (
                        "Cannot start increase data because a task is in ",
                        "progress.\n\nThe current task must be allowed to ",
                        "finish, or be stopped, before starting task.",
                    )
                ),
            )
            return
        if self._import_done:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Increase Data",
                message="".join(
                    (
                        "The import has been done.",
                        "\n\nIncrease data is intended for before import.",
                    )
                ),
            )
            return
        self._allow_job = False
        self._task_name = "increase data"
        self.quit_event.clear()
        self.deferred_update = IncreaseDataProcess(
            self.database,
            self.report_queue,
            self.quit_event,
        )
        self.deferred_update.process.start()
        self.quit_thread = multiprocessing.dummy.DummyProcess(
            target=self._deferred_update_estimate_join
        )
        self.quit_thread.start()

    def _increase_index(self, event=None):
        """Run Increase Index Size process."""
        del event
        if not self._allow_job:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Increase Index",
                message="".join(
                    (
                        "Cannot start increase index because a task is in ",
                        "progress.\n\nThe current task must be allowed to ",
                        "finish, or be stopped, before starting task.",
                    )
                ),
            )
            return
        if self._import_done:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Increase Index",
                message="".join(
                    (
                        "The import has been done.",
                        "\n\nIncrease index is intended for before import.",
                    )
                ),
            )
            return
        self._allow_job = False
        self._task_name = "increase index"
        self.quit_event.clear()
        self.deferred_update = IncreaseIndexProcess(
            self.database,
            self.report_queue,
            self.quit_event,
        )
        self.deferred_update.process.start()
        self.quit_thread = multiprocessing.dummy.DummyProcess(
            target=self._deferred_update_estimate_join
        )
        self.quit_thread.start()

    def _do_import(self, event=None):
        """Run import process if allowed and not already run.

        event is ignored and is present for compatibility between button click
        and keypress.

        """
        del event
        if not self._allow_job:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Import",
                message="".join(
                    (
                        "Cannot start import because a task is in progress",
                        ".\n\nThe current task must be allowed to finish, ",
                        "or be stopped, before starting task.",
                    )
                ),
            )
            return
        if self._import_done:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Import",
                message="".join(
                    (
                        "The import has been done.",
                        "\n\nDismiss Log and start again to repeat it or ",
                        "do another one.",
                    )
                ),
            )
            return
        if not tkinter.messagebox.askokcancel(
            parent=self.root,
            title="Import",
            message="".join(("Please confirm the import is to be started.",)),
        ):
            return
        self._allow_job = False
        self._task_name = "import"
        self.quit_event.clear()
        self.deferred_update = DeferredUpdateProcess(
            self.database,
            self.dumethod,
            self.report_queue,
            self.quit_event,
            self.increases,
            self.home_directory,
            self.pgnfiles,
        )
        self.deferred_update.process.start()
        self.quit_thread = multiprocessing.dummy.DummyProcess(
            target=self._deferred_update_join
        )
        self.quit_thread.start()

    def _stop_task(self, event=None):
        """Stop task.

        event is ignored and is present for compatibility between button click
        and keypress.

        """
        del event
        if self._allow_job:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Stop",
                message="No task running to be stopped.",
            )
            return
        if not tkinter.messagebox.askokcancel(
            parent=self.root,
            title="Stop",
            message=self._task_name.join(
                ("Please confirm the ", " task is to be stopped.")
            ),
        ):
            return
        self.quit_event.set()

    def _dismiss_import_log(self, event=None):
        """Dismiss log display and quit process.

        event is ignored and is present for compatibility between button click
        and keypress.

        """
        del event
        if not self._allow_job:
            tkinter.messagebox.showinfo(
                parent=self.root,
                title="Dismiss",
                message="".join(
                    (
                        "Cannot dismiss because a task is in progress",
                        ".\n\nThe current task must be allowed to finish, ",
                        "or be stopped, before dismissing.",
                    )
                ),
            )
            return
        if tkinter.messagebox.askyesno(
            parent=self.root,
            title="Dismiss",
            message="Do you want to dismiss the import log?",
        ):
            self.root.destroy()



# dptfastload_dir.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import directory of PGN files with dpt.chessdptfastload to database."""


if __name__ == "__main__":
    from .directory_widget import DirectoryWidget
    from .chessdptfastload import chess_dptfastload

    DirectoryWidget(chess_dptfastload, "dpt fastload")



# displaypgn.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which define set bindings and navigation methods.

The methods in these classes are shared by classes in gamedisplay and
repertoiredisplay modules which display Portable Game Notation (PGN) text.

The gamedisplay module has two sets of classes: based on the _GameDisplay and
GameDialogue classes.

The repertoiredisplay module has an identical structure where the _GameDisplay
and _RepertoireDisplay classes have many methods in common.

All methods in the classes in this module existed as multiple copies in various
classes in the gamedisplay and repertoiredisplay modules.  They are now deleted
from those modules.

The classes in this module represent the different sets of classes with methods
in common.  Two classes could reasonably be called DisplayPGN: the choice is
consistent with the naming of InsertPGN and EditPGN.  ShowPGN was chosen for
the other class.

The ShowPGN class is populated with the methods identical in _GameDisplay
and _RepertoireDisplay, which were then removed from those two classes.

The DisplayPGN class is populated with the methods identical in
GameDisplay and RepertoireDisplay, which were then removed
from those two classes.

The InsertPGN class is populated with the methods identical in
GameDisplayInsert and RepertoireDisplayInsert, which were then removed
from those two classes.

The EditPGN class is populated with the methods identical in
GameDisplayEdit and RepertoireDisplayEdit, which were then removed
from those two classes.  It probably should be a subclass of InsertPGN, but
this depends on successful choice of method resolution order in the classes
in gamedisplay and repertoiredisplay modules.

"""
import tkinter
import tkinter.messagebox

from solentware_grid.gui.dataedit import RecordEdit
from solentware_grid.gui.datadelete import RecordDelete

from .score import NonTagBind, ScoreNoGameException
from .scorepgn import ScorePGN
from .displaytext import ShowText, DisplayText, EditText, InsertText


# ShowPGN because DisplayPGN fits GameDisplay (and *Repertoire*)
# ShowText before ScorePGN because identical methods in ShowPGN and ShowText
# are deleted from ShowPGN.
class ShowPGN(ShowText, ScorePGN):
    """Provide methods to set bindinds and traverse visible items."""

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def bind_for_item_navigation(self):
        """Set bindings to navigate PGN score on pointer click."""
        if self.score is self.takefocus_widget:
            super().bind_for_item_navigation()
            self.set_board_pointer_move_bindings(True)
            self.analysis.set_score_pointer_item_navigation_bindings(False)
            self.set_score_pointer_to_score_bindings(False)
            self.set_analysis_score_pointer_to_analysis_score_bindings(True)
        else:
            self.analysis.set_board_pointer_move_bindings(True)
            self.set_score_pointer_item_navigation_bindings(False)
            self.analysis.set_score_pointer_item_navigation_bindings(True)
            self.set_score_pointer_to_score_bindings(True)
            self.set_analysis_score_pointer_to_analysis_score_bindings(False)
        self.set_toggle_game_analysis_bindings(True)

    def bind_for_widget_navigation(self):
        """Set bindings to give focus to this PGN score on pointer click."""
        super().bind_for_widget_navigation()
        self._set_board_pointer_widget_navigation_bindings(True)
        self.set_analysis_score_pointer_to_analysis_score_bindings(False)
        self.set_analysis_score_pointer_to_analysis_score_bindings(False)

    # Probably becomes _set_item(), but stays here rather than moved to each
    # subclass: see displaytext.ShowText version.
    def set_and_tag_item_text(self, reset_undo=False):
        """Delegate to superclass method and set PGN score inactive."""
        # Superclass may set self._most_recent_bindings but test below must be
        # against current value.
        mrb = self._most_recent_bindings

        try:
            super().set_and_tag_item_text(reset_undo=reset_undo)
        except ScoreNoGameException:
            return
        if mrb != NonTagBind.NO_EDITABLE_TAGS:
            for event_spec in (self._get_inactive_button_events(),):
                self.set_event_bindings_score(event_spec, switch=True)

    def _set_select_variation_bindings(self, switch=True):
        """Delegate then set navigation and close item bindings."""
        super()._set_select_variation_bindings(switch=switch)
        self._set_database_navigation_close_item_bindings(switch=switch)

    # The methods identical except for docstrings, and references to
    # self.ui.game_items or self.ui.repertoire_items replaced by property
    # self.ui_displayed_items.

    def _next_item(self, event=None):
        """Select next PGN score on display."""
        if self.ui_displayed_items.count_items_in_stack() > 1:
            self._cycle_item(prior=False)

    def _prior_item(self, event=None):
        """Select previous PGN score on display."""
        if self.ui_displayed_items.count_items_in_stack() > 1:
            self._cycle_item(prior=True)

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        self._set_board_pointer_widget_navigation_bindings(True)
        return super().traverse_backward(event=event)

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        self._set_board_pointer_widget_navigation_bindings(True)
        return super().traverse_forward(event=event)

    def give_focus_to_widget(self, event=None):
        """Select PGN score on display by mouse click."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        losefocus, gainfocus = self.ui_displayed_items.give_focus_to_widget(
            event.widget
        )
        if losefocus is not gainfocus:
            self.ui_configure_item_list_grid()
            gainfocus.set_game_list()
        return "break"

    # The insert_game_database method, coerced into sameness from the methods
    # in gamedisplay._GameDisplay and repertoiredisplay._RepertoireDisplay with
    # class attibutes pgn_score_name, pgn_score_source, pgn_score_tags,
    # and method mark_partial_positions_to_be_recalculated, and property
    # ui_base_table.  The clarity of both common bits and differences
    # seems to justify the extra syntactic complexity.

    # Probably becomes _insert_item_database(), but stays here rather than
    # moved to each subclass: see displaytext.ShowText version.
    def _insert_item_database(self, event=None):
        """Add PGN score to database on request from item display."""
        del event
        title = " ".join(("Insert", self.pgn_score_name.title()))
        psn = self.pgn_score_name
        if self.ui_displayed_items.active_item is None:
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("No active ", " to insert into database.")),
            )
            return None

        # This should see if game with same PGN Tags already exists,
        # after checking for database open, and offer option to insert anyway.
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("Cannot add ", ":\n\nNo database open.")),
            )
            return None

        datasource = self.ui_base_table.get_data_source()
        if datasource is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("Cannot add ", ":\n\n", " list hidden.")),
            )
            return None
        if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
            parent=self.ui.get_toplevel(),
            title=title,
            message=psn.join(("Confirm request to add ", " to database")),
        ):
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("Add ", " to database abandonned.")),
            )
            return None
        updater = self._game_updater(repr(self.score.get("1.0", tkinter.END)))
        if not updater.value.collected_game.is_pgn_valid():
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(
                    (
                        "The new ",
                        "".join(
                            (
                                " score contains at least one illegal move ",
                                "in PGN.\n\nPlease re-confirm request to ",
                                "insert ",
                            )
                        ),
                        ".",
                    )
                ),
            ):
                return None
            updater.value.set_game_source(self.pgn_score_source)
        editor = RecordEdit(updater, None)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        self.mark_partial_positions_to_be_recalculated(datasource=datasource)
        updater.key.recno = None
        editor.put()
        tags = updater.value.collected_game.pgn_tags
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title=title,
            message="".join(
                (
                    psn.title(),
                    ' "',
                    "  ".join([tags.get(k, "") for k in self.pgn_score_tags]),
                    '" added to database.',
                )
            ),
        )
        return True


class DisplayPGN(DisplayText):
    """Provide method to delete an item from database."""

    def _delete_item_database(self, event=None):
        """Remove PGN score from database on request from item display."""
        del event
        title = " ".join(("Delete", self.pgn_score_name.title()))
        psn = self.pgn_score_name
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("Cannot delete ", ":\n\nNo database open.")),
            )
            return
        datasource = self.ui_base_table.get_data_source()
        if datasource is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("Cannot delete ", ":\n\n", " list hidden.")),
            )
            return
        if self.sourceobject is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(
                    (
                        "Cannot delete ",
                        "".join(
                            (
                                ":\n\nDatabase has been closed since ",
                                "this copy displayed.",
                            )
                        ),
                    )
                ),
            )
            return
        if self.blockchange:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(
                    (
                        "Cannot delete ",
                        "".join(
                            (
                                ":\n\nRecord has been amended since ",
                                "this copy displayed.",
                            )
                        ),
                    )
                ),
            )
            return
        if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
            parent=self.ui.get_toplevel(),
            title=title,
            message=psn.join(("Confirm request to delete ", " from database")),
        ):
            return
        original = self.pgn_score_updater()
        original.load_record(
            (self.sourceobject.key.recno, self.sourceobject.srvalue)
        )
        self.pgn_score_original_value(original.value)
        editor = RecordDelete(original)
        editor.set_data_source(datasource, editor.on_data_change)
        self.mark_partial_positions_to_be_recalculated(datasource=datasource)
        editor.delete()
        tags = original.value.collected_game.pgn_tags
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title=title,
            message="".join(
                (
                    psn.title(),
                    ' "',
                    "  ".join([tags.get(k, "") for k in self.pgn_score_tags]),
                    '" deleted from database.',
                )
            ),
        )


class InsertPGN(InsertText):
    """Provide methods to generate popup menus for inserting PGN scores."""

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def _create_primary_activity_popup(self):
        """Delegate then set bindings for navigation and insert PGN."""
        popup = super()._create_primary_activity_popup()
        self._add_pgn_navigation_to_submenu_of_popup(
            popup, index=self.analyse_popup_label
        )
        self._add_pgn_insert_to_submenu_of_popup(
            popup,
            include_ooo=True,
            include_move_rav=True,
            index=self.analyse_popup_label,
        )
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_select_move_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_select_move_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_pgn_tag_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_pgn_tag_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_comment_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_comment_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_nag_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_nag_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_start_rav_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_start_rav_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_end_rav_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_end_rav_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_comment_to_end_of_line_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_comment_to_end_of_line_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_escape_whole_line_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_escape_whole_line_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_reserved_popup(self):
        """Delegate then set bindings for close item."""
        popup = super()._create_reserved_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _set_edit_symbol_mode_bindings(self, switch=True, **ka):
        """Delegate then set bindings for navigation and close item."""
        super()._set_edit_symbol_mode_bindings(switch=switch, **ka)
        self._set_database_navigation_close_item_bindings(switch=switch)


class EditPGN(EditText):
    """Provide methods to generate popup menus for editing PGN scores."""

    def _update_item_database(self, event=None):
        """Modify existing PGN score record."""
        del event
        title = " ".join(("Edit", self.pgn_score_name.title()))
        psn = self.pgn_score_name
        if self.ui.database is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("Cannot edit ", ":\n\nNo database open.")),
            )
            return
        datasource = self.ui_base_table.get_data_source()
        if datasource is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(("Cannot edit ", ":\n\n", " list hidden.")),
            )
            return
        if self.sourceobject is None:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(
                    (
                        "Cannot edit ",
                        "".join(
                            (
                                ":\n\nDatabase has been closed since ",
                                "this copy displayed.",
                            )
                        ),
                    )
                ),
            )
            return
        if self.blockchange:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(
                    (
                        "Cannot edit ",
                        "".join(
                            (
                                ":\n\nRecord has been amended since ",
                                "this copy displayed.",
                            )
                        ),
                    )
                ),
            )
            return
        if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
            parent=self.ui.get_toplevel(),
            title=title,
            message=psn.join(("Confirm request to edit ", ".")),
        ):
            return
        original = self.pgn_score_updater()
        original.load_record(
            (self.sourceobject.key.recno, self.sourceobject.srvalue)
        )
        self.pgn_score_original_value(original.value)

        # is it better to use DataClient directly?
        # Then original would not be used. Instead DataSource.new_row
        # gets record keyed by sourceobject and update is used to edit this.
        text = self.get_score_error_escapes_removed()
        updater = self._game_updater(repr(text))
        editor = RecordEdit(updater, original)
        editor.set_data_source(datasource, editor.on_data_change)
        updater.set_database(editor.get_data_source().dbhome)
        if not updater.value.collected_game.is_pgn_valid():
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.ui.get_toplevel(),
                title=title,
                message=psn.join(
                    (
                        "The edited ",
                        "".join(
                            (
                                " score contains at least one illegal move ",
                                "in PGN.\n\nPlease re-confirm request to ",
                                "edit ",
                            )
                        ),
                        ".",
                    )
                ),
            ):
                return
            updater.value.set_game_source(self.pgn_score_source)
        original.set_database(editor.get_data_source().dbhome)
        updater.key.recno = original.key.recno
        self.mark_partial_positions_to_be_recalculated(datasource=datasource)
        editor.edit()
        if self is self.ui_displayed_items.active_item:
            newkey = self.ui_displayed_items.adjust_edited_item(updater)
            if newkey:
                self._set_properties_on_grids(newkey)
        tags = original.value.collected_game.pgn_tags
        tkinter.messagebox.showinfo(
            parent=self.ui.get_toplevel(),
            title=title,
            message="".join(
                (
                    psn.title(),
                    ' "',
                    "  ".join([tags.get(k, "") for k in self.pgn_score_tags]),
                    '" amended on database.',
                )
            ),
        )



# analysis.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess engine analysis for a position."""

from pgn_read.core.constants import (
    FEN_WHITE_ACTIVE,
    FEN_BLACK_ACTIVE,
)

from .constants import (
    UNKNOWN_RESULT,
    START_RAV,
    END_RAV,
    START_COMMENT,
    END_COMMENT,
    START_EOL_COMMENT,
    FEN_CONTEXT,
)

_EOL_COMMENT_CONTEXT = (START_EOL_COMMENT, "\n")


class Analysis:
    """Chess engine analysis for FEN position.

    Store variations generated by chess engines for a position.

    """

    def __init__(self, position=None):
        """Note position to be analysed."""
        super().__init__()

        # Position encoded as index value for position index.
        self.position = position

        # Variations generated by chess engines:
        # key is engine name reported by chess engine when started,
        # value is ((evaluation, PGN of variation), ...) sorted by evaluation.
        # len(value) is multiPV option value used to do analysis.
        self.variations = {}

        # Depth and width of analysis by chess engine
        # key is engine name reported by chess engine when started,
        # value is (depth, multiPV) option values used to do analysis.
        self.scale = {}

    def set_variations_empty(self):
        """Set all variations to None.

        len(variations[engine] is the multiPV value used when analysis done.
        Set variations to None to preserve the multiPV information when a new
        analysis request is done, while noting any analysis deemed necessary
        is not yet done.

        """
        var = self.variations
        for key, value in var.items():
            var[key] = [None] * len(value)

    # Rewrite of pgn_read in 2020 forced inclusion of SetUp tag in analysis
    # PGN text: probably best to find a way of avoiding this.
    def translate_analysis_to_pgn(self, move_played=""):
        """Translate UCI chess engine variation output to PGN."""
        # Addition of this code tips balance in favour of a class to deal with
        # FEN representation of positions.
        to_move = self.position.split()[1]

        if move_played:
            move_played = "".join(
                (
                    move_played,
                    " ;Move played",
                    " by white"
                    if to_move == FEN_WHITE_ACTIVE
                    else " by black"
                    if to_move == FEN_BLACK_ACTIVE
                    else "",
                    "\n",
                )
            )
        variations = self.variations
        scale = self.scale
        new_text = []
        for engine_name in sorted(variations):
            # The following stopped happening after the test database was
            # deleted and re-created, so maybe everything is sound and the
            # problem is left-over from earlier testing.
            #
            # This may happen if communication with a remote engine is lost
            # and then re-established by 'quit engine' and 'start engine' menu
            # actions, or perhaps if the process driving a local engine fails
            # too.  It has been seen only for the position on display in the
            # active window at the time these actions are done while the
            # database is open.
            # The sequence of actions is:
            # Start a remote UCI server and open a database in either order.
            # Use the Engine menu to start a UCI client for the UCI server.
            # Close the database (so no analysis get written to database).
            # Use the Game menu to start a new game and see analysis appear.
            # Stop the remote UCI server (action on the remote host).
            # Start a remote UCI server.
            # Open a database.
            # Use the Engine menu to start a UCI client for the UCI server.
            # Now closing and opening the database switches it off and on.
            # if engine_name is None:
            #    sys.stderr.write(
            #        'Name of engine: ' + str(engine_name) \
            #        + ' ' + str(len(variations)) + '\n')
            #    continue

            analysis = variations[engine_name]

            # Keep going as normal for non-str engine_name (should not happen).
            # new_text.append(engine_name.join((';', '\n')))
            new_text.append(str(engine_name).join(_EOL_COMMENT_CONTEXT))

            depth, multipv = [str(s) for s in scale[engine_name]]
            # pylint message unused-variable.
            # multipv documents what the second element is..
            del multipv
            lines = [
                "".join(
                    (
                        " ".join(
                            (
                                START_COMMENT,
                                self._evalution_score(a[0], to_move),
                                depth,
                                str(e + 1),
                                " ".join((END_COMMENT, START_RAV)),
                            )
                        ),
                        a[1],
                        END_RAV,
                        "\n",
                    )
                )
                for e, a in enumerate(analysis)
            ]
            new_text.extend(lines)
            if not move_played:
                move_played = "".join(
                    (
                        analysis[0][1].split()[0],
                        " ",
                        "".join(
                            (
                                "First variation",
                                ", white to move"
                                if to_move == FEN_WHITE_ACTIVE
                                else ", black to move"
                                if to_move == FEN_BLACK_ACTIVE
                                else "",
                            )
                        ).join(_EOL_COMMENT_CONTEXT),
                    )
                )
        new_text.append(UNKNOWN_RESULT)
        if move_played:
            new_text.insert(0, move_played)
            new_text.insert(0, self.position.join(FEN_CONTEXT))
        return "".join(new_text)

    @staticmethod
    def _evalution_score(val, to_move):
        """Normalize to white advantage is positive and black negative."""
        return "{:+.2f}".format(
            (int(val) if to_move != FEN_BLACK_ACTIVE else -int(val)) / 100
        )



# filespec.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Files and fields for chess database.

Specification for sqlite3 database, Berkeley DB database and DPT database.

"""

from solentware_base.core.constants import (
    PRIMARY,
    SECONDARY,
    BTOD_FACTOR,
    DEFAULT_RECORDS,
    DEFAULT_INCREASE_FACTOR,
    BTOD_CONSTANT,
    DDNAME,
    FILE,
    FIELDS,
    FILEDESC,
    INV,
    ORD,
    RRN,
    BRECPPG,
    FILEORG,
    DPT_PRIMARY_FIELD_LENGTH,
)
import solentware_base.core.filespec

# Reference profile of a typical game (see FileSpec docstring)
POSITIONS_PER_GAME = 75
PIECES_PER_POSITION = 23
PIECES_TYPES_PER_POSITION = 10  # Added after introduction of chessql.
BYTES_PER_GAME = 700  # Records per page set to 10 because 700 was good.

# Names used to refer to file descriptions
# DPT files or Berkeley DB primary databases
GAMES_FILE_DEF = "games"
PARTIAL_FILE_DEF = "partial"
REPERTOIRE_FILE_DEF = "repertoire"
ANALYSIS_FILE_DEF = "analysis"
SELECTION_FILE_DEF = "selection"
ENGINE_FILE_DEF = "engine"

# Names used to refer to field descriptions
# DPT fields or Berkeley DB secondary databases
GAME_FIELD_DEF = "Game"
SOURCE_FIELD_DEF = "source"
EVENT_FIELD_DEF = "Event"
SITE_FIELD_DEF = "Site"
DATE_FIELD_DEF = "Date"
ROUND_FIELD_DEF = "Round"
WHITE_FIELD_DEF = "White"
BLACK_FIELD_DEF = "Black"
RESULT_FIELD_DEF = "Result"
POSITIONS_FIELD_DEF = "positions"
PIECESQUAREMOVE_FIELD_DEF = "piecesquaremove"
PIECEMOVE_FIELD_DEF = "piecemove"
SQUAREMOVE_FIELD_DEF = "squaremove"
PARTIAL_FIELD_DEF = "Partial"
PARTIALPOSITION_FIELD_DEF = "partialposition"
PARTIALPOSITION_NAME_FIELD_DEF = "partialpositionname"
NEWGAMES_FIELD_DEF = "newgames"
REPERTOIRE_FIELD_DEF = "Repertoire"
OPENING_FIELD_DEF = "Opening"
OPENING_ERROR_FIELD_DEF = "openingerror"
PGN_DATE_FIELD_DEF = "pgndate"
ANALYSIS_FIELD_DEF = "Analysis"
VARIATION_FIELD_DEF = "variation"
ENGINE_FIELD_DEF = "engine"
SELECTION_FIELD_DEF = "rulename"
RULE_FIELD_DEF = "rule"
PROGRAM_FIELD_DEF = "program"
COMMAND_FIELD_DEF = "command"

# Non-standard field names. Standard is x_FIELD_DEF.title()
_PIECESQUAREMOVE_FIELD_NAME = "PieceSquareMove"
_PIECEMOVE_FIELD_NAME = "PieceMove"
_SQUAREMOVE_FIELD_NAME = "SquareMove"
_PARTIALPOSITION_FIELD_NAME = "PartialPosition"
_PP_NAME_FIELD_NAME = "PartialPositionName"
_NEWGAMES_FIELD_NAME = "NewGames"
_OPENING_ERROR_FIELD_NAME = "OpeningError"
_PGN_DATE_FIELD_NAME = "PGNdate"

# Berkeley DB environment.
DB_ENVIRONMENT_GIGABYTES = 0
DB_ENVIRONMENT_BYTES = 1024000
DB_ENVIRONMENT_MAXLOCKS = 120000  # OpenBSD only.
DB_ENVIRONMENT_MAXOBJECTS = 120000  # OpenBSD only.

# Symas LMMD environment.
LMMD_MINIMUM_FREE_PAGES_AT_START = 20000

# Any partial position indexed by NEWGAMES_FIELD_VALUE on _NEWGAMES_FIELD_NAME
# has not been recalculated since an update to the games file.  The partial
# position's _PARTIALGAMES_FIELD_NAME reference on the games file is out of
# date and may be wrong.
NEWGAMES_FIELD_VALUE = "changed"


class FileSpec(solentware_base.core.filespec.FileSpec):
    """Specify a chess database.

    Parameters for Berkeley DB, DPT, and Sqlite3, are defined.

    Ignore settings irrelevant to the database engine in use.

    Berkeley DB and Sqlite3 look after themselves for sizing purposes when
    increasing the size of a database.

    The sizing parameters for the Games file in DPT were chosen using figures
    from a few game collections: ~25000 games in 4NCL up to 2011, ~4500 games
    in a California collection spanning the 20th century, ~2500 games in a
    collection of Volga Gambit games, ~150000 games from section 00 of
    enormous.pgn, ~500 games from 1997 4NCL divisions one and two separately.

             <            per game              >  <     pages    >
      Games Positions Pieces Data bytes PGN Bytes  Table B  Table D   Ratio
      25572     74      23      493        761       2557     20148     7.9
       4445     75      23      418        682        445      6034    13.6
       2324     76      23      416        683        232      3308    14.3
     156954     75      23      375        637      15803    120932     7.7
        528     77      23      481        757         52       891    17.1
        528     74      23      472        739         52       863    16.6

    There is not enough data to conclude that Table D usage is proportional
    to the index entries per game, but it is reasonable to expect this and
    the Positions and Table D ratios in the two 528 game samples are close
    enough to make the assumption.  Above, say, 20000 games it is assumed
    that typical games, 75 positions per game and 23 pieces per position,
    lead to a Table B to Table D ratio of 8, or 14 when CQL statements are
    supported.  The higher ratios at lower numbers of games are assumed to
    arise from the lower chance of compressed inverted lists.

    It is completely unclear what value to use for records per Table B page
    because games can have comments and variations as well as just the moves
    played.  The samples have few comments or variations. A value of 10 seems
    as good a compromise as any between wasting space and getting all record
    numbers used. 16 average sized records would fit on a Table B page.

    """

    def __init__(
        self, use_specification_items=None, dpt_records=None, **kargs
    ):
        """Define chess database."""
        dptdsn = FileSpec.dpt_dsn
        field_name = FileSpec.field_name

        super().__init__(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
            **{
                GAMES_FILE_DEF: {
                    DDNAME: "GAMES",
                    FILE: dptdsn(GAMES_FILE_DEF),
                    FILEDESC: {
                        BRECPPG: 10,
                        FILEORG: RRN,
                    },
                    BTOD_FACTOR: 14,
                    BTOD_CONSTANT: 800,
                    DEFAULT_RECORDS: 10000,
                    DEFAULT_INCREASE_FACTOR: 0.01,
                    PRIMARY: field_name(GAME_FIELD_DEF),
                    DPT_PRIMARY_FIELD_LENGTH: 200,
                    SECONDARY: {
                        SOURCE_FIELD_DEF: SOURCE_FIELD_DEF.title(),
                        EVENT_FIELD_DEF: None,
                        SITE_FIELD_DEF: None,
                        DATE_FIELD_DEF: None,
                        ROUND_FIELD_DEF: None,
                        WHITE_FIELD_DEF: None,
                        BLACK_FIELD_DEF: None,
                        RESULT_FIELD_DEF: None,
                        POSITIONS_FIELD_DEF: POSITIONS_FIELD_DEF.title(),
                        PIECESQUAREMOVE_FIELD_DEF: _PIECESQUAREMOVE_FIELD_NAME,
                        PIECEMOVE_FIELD_DEF: _PIECEMOVE_FIELD_NAME,
                        SQUAREMOVE_FIELD_DEF: _SQUAREMOVE_FIELD_NAME,
                        PARTIALPOSITION_FIELD_DEF: _PARTIALPOSITION_FIELD_NAME,
                        PGN_DATE_FIELD_DEF: _PGN_DATE_FIELD_NAME,
                    },
                    FIELDS: {
                        field_name(GAME_FIELD_DEF): None,
                        SOURCE_FIELD_DEF.title(): {INV: True, ORD: True},
                        WHITE_FIELD_DEF: {INV: True, ORD: True},
                        BLACK_FIELD_DEF: {INV: True, ORD: True},
                        EVENT_FIELD_DEF: {INV: True, ORD: True},
                        ROUND_FIELD_DEF: {INV: True, ORD: True},
                        DATE_FIELD_DEF: {INV: True, ORD: True},
                        RESULT_FIELD_DEF: {INV: True, ORD: True},
                        SITE_FIELD_DEF: {INV: True, ORD: True},
                        POSITIONS_FIELD_DEF.title(): {
                            INV: True,
                            ORD: True,
                        },  # , ACCESS_METHOD:HASH},
                        _PIECESQUAREMOVE_FIELD_NAME: {
                            INV: True,
                            ORD: True,
                        },  # , ACCESS_METHOD:HASH},
                        _PIECEMOVE_FIELD_NAME: {
                            INV: True,
                            ORD: True,
                        },  # , ACCESS_METHOD:HASH},
                        _SQUAREMOVE_FIELD_NAME: {
                            INV: True,
                            ORD: True,
                        },  # , ACCESS_METHOD:HASH},
                        _PARTIALPOSITION_FIELD_NAME: {INV: True, ORD: True},
                        _PGN_DATE_FIELD_NAME: {INV: True, ORD: True},
                    },
                },
                PARTIAL_FILE_DEF: {
                    DDNAME: "PARTIAL",
                    FILE: dptdsn(PARTIAL_FILE_DEF),
                    FILEDESC: {
                        BRECPPG: 40,
                        FILEORG: RRN,
                    },
                    BTOD_FACTOR: 1,  # a guess
                    BTOD_CONSTANT: 100,  # a guess
                    DEFAULT_RECORDS: 10000,
                    DEFAULT_INCREASE_FACTOR: 0.5,
                    PRIMARY: field_name(PARTIAL_FIELD_DEF),
                    DPT_PRIMARY_FIELD_LENGTH: 127,
                    SECONDARY: {
                        PARTIALPOSITION_NAME_FIELD_DEF: _PP_NAME_FIELD_NAME,
                        NEWGAMES_FIELD_DEF: _NEWGAMES_FIELD_NAME,
                    },
                    FIELDS: {
                        field_name(PARTIAL_FIELD_DEF): None,
                        _PP_NAME_FIELD_NAME: {INV: True, ORD: True},
                        _NEWGAMES_FIELD_NAME: {INV: True, ORD: True},
                    },
                },
                REPERTOIRE_FILE_DEF: {
                    DDNAME: "REPERT",
                    FILE: dptdsn(REPERTOIRE_FILE_DEF),
                    FILEDESC: {
                        BRECPPG: 1,
                        FILEORG: RRN,
                    },
                    BTOD_FACTOR: 0.1,
                    BTOD_CONSTANT: 800,
                    DEFAULT_RECORDS: 100,
                    DEFAULT_INCREASE_FACTOR: 0.01,
                    PRIMARY: field_name(REPERTOIRE_FIELD_DEF),
                    DPT_PRIMARY_FIELD_LENGTH: 200,
                    SECONDARY: {
                        OPENING_FIELD_DEF: None,
                        OPENING_ERROR_FIELD_DEF: _OPENING_ERROR_FIELD_NAME,
                    },
                    FIELDS: {
                        field_name(REPERTOIRE_FIELD_DEF): None,
                        OPENING_FIELD_DEF: {INV: True, ORD: True},
                        _OPENING_ERROR_FIELD_NAME: {INV: True, ORD: True},
                    },
                },
                ANALYSIS_FILE_DEF: {
                    DDNAME: "ANALYSIS",
                    FILE: dptdsn(ANALYSIS_FILE_DEF),
                    FILEDESC: {
                        BRECPPG: 10,
                        FILEORG: RRN,
                    },
                    BTOD_FACTOR: 1,
                    BTOD_CONSTANT: 800,
                    DEFAULT_RECORDS: 100000,
                    DEFAULT_INCREASE_FACTOR: 1.0,
                    PRIMARY: field_name(ANALYSIS_FIELD_DEF),
                    DPT_PRIMARY_FIELD_LENGTH: 200,
                    SECONDARY: {
                        ENGINE_FIELD_DEF: ENGINE_FIELD_DEF.title(),
                        VARIATION_FIELD_DEF: VARIATION_FIELD_DEF.title(),
                    },
                    FIELDS: {
                        field_name(ANALYSIS_FIELD_DEF): None,
                        ENGINE_FIELD_DEF.title(): {INV: True, ORD: True},
                        VARIATION_FIELD_DEF.title(): {INV: True, ORD: True},
                    },
                },
                SELECTION_FILE_DEF: {
                    DDNAME: "SLCTRULE",
                    FILE: dptdsn(SELECTION_FILE_DEF),
                    FILEDESC: {
                        BRECPPG: 20,
                        FILEORG: RRN,
                    },
                    BTOD_FACTOR: 1,  # a guess
                    BTOD_CONSTANT: 100,  # a guess
                    DEFAULT_RECORDS: 10000,
                    DEFAULT_INCREASE_FACTOR: 0.5,
                    PRIMARY: field_name(SELECTION_FIELD_DEF),
                    DPT_PRIMARY_FIELD_LENGTH: 127,
                    SECONDARY: {
                        RULE_FIELD_DEF: RULE_FIELD_DEF.title(),
                    },
                    FIELDS: {
                        field_name(SELECTION_FIELD_DEF): None,
                        RULE_FIELD_DEF.title(): {INV: True, ORD: True},
                    },
                },
                ENGINE_FILE_DEF: {
                    DDNAME: "ENGINE",
                    FILE: dptdsn(ENGINE_FILE_DEF),
                    FILEDESC: {
                        BRECPPG: 150,
                        FILEORG: RRN,
                    },
                    BTOD_FACTOR: 1,  # a guess
                    BTOD_CONSTANT: 100,  # a guess
                    DEFAULT_RECORDS: 1000,
                    DEFAULT_INCREASE_FACTOR: 0.5,
                    PRIMARY: field_name(PROGRAM_FIELD_DEF),
                    DPT_PRIMARY_FIELD_LENGTH: 127,
                    SECONDARY: {
                        COMMAND_FIELD_DEF: COMMAND_FIELD_DEF.title(),
                    },
                    FIELDS: {
                        field_name(PROGRAM_FIELD_DEF): None,
                        COMMAND_FIELD_DEF.title(): {INV: True, ORD: True},
                    },
                },
            }
        )



# unqlitedu_file.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with unqlite.chessunqlitedu to database."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..unqlite.chessunqlitedu import ChessDatabase

    FileWidget(ChessDatabase, "unqlite")



# file_widget_fastunload.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Try fast unload of a ChessTab database.

Purpose is to verify fast unload functions correctly for at least one
ChessTab database, and generate sets of 'TAPE' files to test fast load.

The attempt to write a fast load input files generator for PGN files from
the description of file formats in DPT DBA Guide failed: so see if there
is a way fast load works and copy that.
"""

import tkinter
import os
import time

import tkinter.messagebox
import tkinter.filedialog

import dptdb.dptapi

from ..dpt.chessdpt import ChessDatabase


def file_fastunload(dbpath, outputdir):
    """Open database, fast unload Games file, and close database."""
    print(time.ctime())
    cdb = ChessDatabase(dbpath, allowcreate=True)
    cdb.open_database()
    for table in cdb.table.values():
        # Unload accepts positional arguments only.
        # Want 'dir' argument as 'dbpath', not the default '#FASTIO' via
        # definition of FUNLOAD_DIR.
        # So have to specify options where FUNLOAD_DEFAULT, itself defined
        # via FUNLOAD_ALLINFO (at time of writing) which is required option,
        # is the default option.
        table.opencontext.Unload(
            dptdb.dptapi.FUNLOAD_DEFAULT, None, None, outputdir
        )
    cdb.close_database()
    print(time.ctime())


class FileWidget:
    """Select ChessTab database to fast unload."""

    def __init__(self):
        """Import games into database using engine_name database engine."""
        root = tkinter.Tk()
        root.wm_title("Fast Unload ChessTab Database")
        root.wm_iconify()
        dbdir = tkinter.filedialog.askdirectory(
            title="Fast Unload ChessTab Database"
        )
        if dbdir:
            outputdir = tkinter.filedialog.askdirectory(
                title="Fast Unload Output Directory"
            )
            if outputdir:
                if tkinter.messagebox.askyesno(
                    title="Fast Unload Games",
                    message="Proceed with Fast Unload",
                ):
                    file_fastunload(dbdir, outputdir)
        root.destroy()


if __name__ == "__main__":
    FileWidget()



# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""View a database of chess games created from data in PGN format.

Run "python -m chesstab.chessgames" assuming chesstab is in site-packages and
Python3.3 or later is the system Python.

PGN is "Portable Game Notation", the standard non-proprietary format for files
of chess game scores.

Sqlite3 via the apsw or sqlite packages, Berkeley DB via the db package, or DPT
via the dpt package, can be used as the database engine.

When importing games while running under Wine it will probably be necessary to
use the module "chessgames_winedptchunk".  The only known reason to run under
Wine is to use the DPT database engine on a platform other than Microsoft
Windows.
"""

from solentware_base.core.constants import (
    BERKELEYDB_MODULE,
    BSDDB3_MODULE,
    DPT_MODULE,
    SQLITE3_MODULE,
    APSW_MODULE,
    UNQLITE_MODULE,
    VEDIS_MODULE,
    GNU_MODULE,
    NDBM_MODULE,
    DB_TCL_MODULE,
    LMDB_MODULE,
)

APPLICATION_NAME = "ChessTab"
ERROR_LOG = "ErrorLog"

# berkeleydb interface module name
_BERKELEYDBCHESS = __name__ + ".berkeleydb.chessberkeleydb"

# bsddb3 interface module name
_DBCHESS = __name__ + ".db.chessdb"

# DPT interface module name
_DPTCHESS = __name__ + ".dpt.chessdpt"

# sqlite3 interface module name
_SQLITE3CHESS = __name__ + ".sqlite.chesssqlite3"

# apsw interface module name
_APSWCHESS = __name__ + ".apsw.chessapsw"

# unqlite interface module name
_UNQLITECHESS = __name__ + ".unqlite.chessunqlite"

# vedis interface module name
_VEDISCHESS = __name__ + ".vedis.chessvedis"

# dbm.gnu interface module name
_GNUCHESS = __name__ + ".gnu.chessgnu"

# dbm.ndbm interface module name
_NDBMCHESS = __name__ + ".ndbm.chessndbm"

# tkinter.tcl Berkeley DB interface module name
_DB_TCLCHESS = __name__ + ".db_tkinter.chessdbtkinter"

# lmdb interface module name
_LMDBCHESS = __name__ + ".lmdb.chesslmdb"

# Map database module names to application module
APPLICATION_DATABASE_MODULE = {
    BERKELEYDB_MODULE: _BERKELEYDBCHESS,
    BSDDB3_MODULE: _DBCHESS,
    SQLITE3_MODULE: _SQLITE3CHESS,
    APSW_MODULE: _APSWCHESS,
    DPT_MODULE: _DPTCHESS,
    UNQLITE_MODULE: _UNQLITECHESS,
    VEDIS_MODULE: _VEDISCHESS,
    GNU_MODULE: _GNUCHESS,
    NDBM_MODULE: _NDBMCHESS,
    DB_TCL_MODULE: _DB_TCLCHESS,
    LMDB_MODULE: _LMDBCHESS,
}

# Default partial position dataset module name
_DEFAULTPARTIALPOSITION = __name__ + ".basecore.cqlds"

# DPT partial position dataset module name
_DPTPARTIALPOSITION = __name__ + ".dpt.cqlds"

# Map database module names to partial position dataset module
PARTIAL_POSITION_MODULE = {
    BERKELEYDB_MODULE: _DEFAULTPARTIALPOSITION,
    BSDDB3_MODULE: _DEFAULTPARTIALPOSITION,
    SQLITE3_MODULE: _DEFAULTPARTIALPOSITION,
    APSW_MODULE: _DEFAULTPARTIALPOSITION,
    DPT_MODULE: _DPTPARTIALPOSITION,
    UNQLITE_MODULE: _DEFAULTPARTIALPOSITION,
    VEDIS_MODULE: _DEFAULTPARTIALPOSITION,
    GNU_MODULE: _DEFAULTPARTIALPOSITION,
    NDBM_MODULE: _DEFAULTPARTIALPOSITION,
    DB_TCL_MODULE: _DEFAULTPARTIALPOSITION,
    LMDB_MODULE: _DEFAULTPARTIALPOSITION,
}

# Default full position dataset module name
_DEFAULTFULLPOSITION = __name__ + ".basecore.fullpositionds"

# DPT full dataset module name
_DPTFULLPOSITION = __name__ + ".dpt.fullpositionds"

# Map database module names to full position dataset module
FULL_POSITION_MODULE = {
    BERKELEYDB_MODULE: _DEFAULTFULLPOSITION,
    BSDDB3_MODULE: _DEFAULTFULLPOSITION,
    SQLITE3_MODULE: _DEFAULTFULLPOSITION,
    APSW_MODULE: _DEFAULTFULLPOSITION,
    DPT_MODULE: _DPTFULLPOSITION,
    UNQLITE_MODULE: _DEFAULTFULLPOSITION,
    VEDIS_MODULE: _DEFAULTFULLPOSITION,
    GNU_MODULE: _DEFAULTFULLPOSITION,
    NDBM_MODULE: _DEFAULTFULLPOSITION,
    DB_TCL_MODULE: _DEFAULTFULLPOSITION,
    LMDB_MODULE: _DEFAULTFULLPOSITION,
}

# Default analysis dataset module name
_DEFAULTANALYSIS = __name__ + ".basecore.analysisds"

# DPT analysis dataset module name
_DPTANALYSIS = __name__ + ".dpt.analysisds"

# Map database module names to analysis dataset module
ANALYSIS_MODULE = {
    BERKELEYDB_MODULE: _DEFAULTANALYSIS,
    BSDDB3_MODULE: _DEFAULTANALYSIS,
    SQLITE3_MODULE: _DEFAULTANALYSIS,
    APSW_MODULE: _DEFAULTANALYSIS,
    DPT_MODULE: _DPTANALYSIS,
    UNQLITE_MODULE: _DEFAULTANALYSIS,
    VEDIS_MODULE: _DEFAULTANALYSIS,
    GNU_MODULE: _DEFAULTANALYSIS,
    NDBM_MODULE: _DEFAULTANALYSIS,
    DB_TCL_MODULE: _DEFAULTANALYSIS,
    LMDB_MODULE: _DEFAULTANALYSIS,
}

# Default selection rules dataset module name
_DEFAULTSELECTION = __name__ + ".basecore.selectionds"

# DPT selection rules dataset module name
_DPTSELECTION = __name__ + ".dpt.selectionds"

# Map database module names to selection rules dataset module
SELECTION_MODULE = {
    BERKELEYDB_MODULE: _DEFAULTSELECTION,
    BSDDB3_MODULE: _DEFAULTSELECTION,
    SQLITE3_MODULE: _DEFAULTSELECTION,
    APSW_MODULE: _DEFAULTSELECTION,
    DPT_MODULE: _DPTSELECTION,
    UNQLITE_MODULE: _DEFAULTSELECTION,
    VEDIS_MODULE: _DEFAULTSELECTION,
    GNU_MODULE: _DEFAULTSELECTION,
    NDBM_MODULE: _DEFAULTSELECTION,
    DB_TCL_MODULE: _DEFAULTSELECTION,
    LMDB_MODULE: _DEFAULTSELECTION,
}



# chessdptdu.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using DPT single-step deferred update.

This module on Windows only.  Use multi-step module on Wine because Wine
support for a critical function used by single-step is not reliable. There
is no sure way to spot that module is running on Wine.

See www.dptoolkit.com for details of DPT

"""
import os
import multiprocessing
import sys

# pylint will always give import-error message on non-Microsoft Windows
# systems.
# Wine counts as a Microsft Windows system.
# It is reasonable to not install 'dptdb.dptapi'.
# The importlib module is used to import chessdptdu if needed.
from dptdb.dptapi import (
    FISTAT_DEFERRED_UPDATES,
    FISTAT_PHYS_BROKEN,
    FIFLAGS_FULL_TABLEB,
    FIFLAGS_FULL_TABLED,
    FLOAD_DEFAULT,
)

from solentware_base import dptdu_database
from solentware_base.core.archivedudpt import ArchiveduDPT
from solentware_base.core.constants import (
    FILEDESC,
    BRECPPG,
    TABLE_B_SIZE,
    DPT_SYSFL_FOLDER,
    BTOD_FACTOR,
)

from .filespec import FileSpec
from ..core.filespec import (
    GAMES_FILE_DEF,
    PIECES_PER_POSITION,
    POSITIONS_PER_GAME,
    PIECES_TYPES_PER_POSITION,
    BYTES_PER_GAME,
)
from ..shared.alldu import (
    chess_du_import,
    chess_du_backup_before_import,
    chess_du_delete_backup_after_import,
)
from . import chessdptnofistat

# The DPT segment size is 65280 because 32 bytes are reserved and 8160 bytes of
# the 8192 byte page are used for the bitmap.
# TABLE_B_SIZE value is necessarily the same byte size, and already defined.
_DEFERRED_UPDATE_POINTS = (TABLE_B_SIZE * 8 - 1,)
del TABLE_B_SIZE


class DPTFileSpecError(Exception):
    """File definition problem in ChessDatabase initialisation."""


class DPTFistatError(Exception):
    """Attempt to open a file when not in deferred update mode."""


class DPTSizingError(Exception):
    """Unable to plan file size increases from PGN import estimates."""


def chess_database_du(
    dbpath, *args, file=None, reporter=None, increases=None, **kwargs
):
    """Open database, import games and close database."""
    files = (file,) if file is not None else None

    # Take a backup of games file by fast dump.
    bu_process = multiprocessing.Process(
        target=chess_database_import_backup,
        args=(dbpath, files),
        kwargs=dict(file=file, reporter=reporter, increases=increases),
    )
    bu_process.start()
    bu_process.join()
    if bu_process.exitcode != 0:
        return
    del bu_process

    # Import to games file.
    import_process = multiprocessing.Process(
        target=chess_database_import,
        args=(dbpath, files, *args),
        kwargs=dict(
            file=file, reporter=reporter, increases=increases, **kwargs
        ),
    )
    import_process.start()
    import_process.join()
    if import_process.exitcode != 0:
        return
    del import_process

    # Recover games file from backup if import failed.
    recover_process = multiprocessing.Process(
        target=chess_database_status_after_import,
        args=(dbpath, files),
        kwargs=dict(
            file=file, reporter=reporter, increases=increases, **kwargs
        ),
    )
    recover_process.start()
    recover_process.join()
    if recover_process.exitcode != 0:
        return
    del recover_process

    # Assume the import succeeded if backup is not present.
    if not os.path.exists(
        os.path.join(dbpath, dptdu_database.Database.import_backup_directory)
    ):
        return

    # Assume recovery from failed import also failed if games file is broken.
    if _games_file_is_broken(dbpath, files, file=file):
        if reporter is not None:
            reporter.append_text("Import not done.")
            reporter.append_text_only("")
        return

    # Try the import once more: table B and table D increases already applied.
    import_process = multiprocessing.Process(
        target=chess_database_import,
        args=(dbpath, files, *args),
        kwargs=dict(file=file, reporter=reporter, increases=None, **kwargs),
    )
    import_process.start()
    import_process.join()
    if import_process.exitcode != 0:
        return
    del import_process

    # Recover games file from backup if re-tried import failed.
    recover_process = multiprocessing.Process(
        target=chess_database_status_after_import,
        args=(dbpath, files),
        kwargs=dict(
            file=file, reporter=reporter, increases=increases, **kwargs
        ),
    )
    recover_process.start()
    recover_process.join()
    del recover_process

    # Assume the import succeeded if backup is not present.
    if not os.path.exists(
        os.path.join(dbpath, dptdu_database.Database.import_backup_directory)
    ):
        return

    # Does not matter how recovery from re-tried import failure turns out.
    if reporter is not None:
        reporter.append_text_only("")
        reporter.append_text("Import not done.")
        reporter.append_text_only("")


def _games_file_is_broken(dbpath, files, file=None):
    """Return True if games file is broken, False otherwise."""
    fistat_process = multiprocessing.Process(
        target=chess_database_current_status,
        args=(dbpath, files),
        kwargs=dict(file=file),
    )
    fistat_process.start()
    fistat_process.join()
    return fistat_process.exitcode != 0


class ChessDatabase(dptdu_database.Database, ArchiveduDPT):
    """Provide deferred update methods for a database of games of chess.

    Subclasses must include a subclass of dptbase.Database as a superclass.

    """

    # ChessDatabase.deferred_update_points is not needed in DPT, like
    # the similar attribute in chessdbbitdu.ChessDatabase for example, because
    # DPT does it's own memory management for deferred updates.
    # The same attribute is provided to allow the import_pgn method called in
    # this module's chess_database_du function to report progress at regular
    # intervals.
    # The values are set differently because Wine does not give a useful answer
    # to DPT's memory usage questions.
    deferred_update_points = frozenset(_DEFERRED_UPDATE_POINTS)

    def __init__(
        self,
        databasefolder,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        ddnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )
        # Deferred update for games file only
        for ddname in list(ddnames.keys()):
            if ddname != GAMES_FILE_DEF:
                del ddnames[ddname]

        if not kargs.get("allowcreate", False):
            try:
                for ddname in ddnames:
                    if FILEDESC in ddnames[ddname]:
                        del ddnames[ddname][FILEDESC]
            except Exception as error:
                if __name__ == "__main__":
                    raise
                raise DPTFileSpecError("DPT description invalid") from error

        try:
            super().__init__(ddnames, databasefolder, **kargs)
        except Exception as error:
            if __name__ == "__main__":
                raise
            raise DPTFileSpecError("DPT description invalid") from error

        # Retain import estimates for increase size by button actions
        self._import_estimates = None
        self._notional_record_counts = None
        # Methods passed by UI to populate report widgets
        self._reporter = None

    def open_database(self, files=None):
        """Delegate then raise DPTFistatError if database not in DU mode.

        Normally return None with the database open.

        Close the database and raise DPTFistatError exception if the
        database FISTAT parameter is not equal FISTAT_DEFERRED_UPDATES.

        """
        super().open_database(files=files)
        viewer = self.dbenv.Core().GetViewerResetter()
        for dbo in self.table.values():
            if (
                viewer.ViewAsInt("FISTAT", dbo.opencontext)
                != FISTAT_DEFERRED_UPDATES
            ):
                break
        else:
            # Previous algorithm called self.increase_database_size here.
            # Now the increase is done in chess_du_import call from
            # chess_database_du function.
            return
        self.close_database()
        raise DPTFistatError("A file is not in deferred update mode")

    def open_context_prepare_import(self, files=None):
        """Open all files normally."""
        super().open_database(files=files)

    def get_pages_for_record_counts(self, counts=(0, 0)):
        """Return Table B and Table D pages needed for record counts."""
        brecppg = self.table[GAMES_FILE_DEF].filedesc[BRECPPG]
        return (
            counts[0] // brecppg,
            (counts[1] * self.table[GAMES_FILE_DEF].btod_factor) // brecppg,
        )

    def _get_database_table_sizes(self, files=None):
        """Return Table B and D size and usage in pages for files."""
        if files is None:
            files = {}
        filesize = {}
        for key, value in self.get_database_parameters(
            files=list(files.keys())
        ).items():
            filesize[key] = (
                value["BSIZE"],
                value["BHIGHPG"],
                value["DSIZE"],
                value["DPGSUSED"],
            )
        increase = self.get_database_increase(files=files)
        self.close_database_contexts()
        return filesize, increase

    def get_file_sizes(self):
        """Return dictionary of notional record counts for data and index."""
        return self._notional_record_counts

    def report_plans_for_estimate(self, estimates, reporter, increases):
        """Calculate and report file size adjustments to do import.

        Note the reporter and headline methods for initial report and possible
        later recalculations.

        Pass estimates through to self._report_plans_for_estimate

        """
        # See comment near end of class definition Chess in relative module
        # ..gui.chess for explanation of this change.
        self._reporter = reporter
        try:
            self._report_plans_for_estimate(
                estimates=estimates,
                increases=increases,
            )
        except DPTSizingError:
            if reporter:
                reporter.append_text_only("")
                reporter.append_text(
                    "No estimates available to calculate file size increase."
                )
        reporter.append_text_only("")
        reporter.append_text("Ready to start import.")

    def _report_plans_for_estimate(self, estimates=None, increases=None):
        """Recalculate and report file size adjustments to do import.

        Create dictionary of effective game counts for sizing Games file.
        This will be passed to the import job which will increase Table B and
        Table D according to file specification.

        The counts for Table B and Table D can be different.  If the average
        data bytes per game is greater than Page size / Records per page the
        count must be increased to allow for the unused record numbers.  If
        the average positions per game or pieces per position are not the
        values used to calculate the steady-state ratio of Table B to Table D
        the count must be adjusted to compensate.

        """
        append_text = self._reporter.append_text
        append_text_only = self._reporter.append_text_only
        if estimates is not None:
            self._import_estimates = estimates
        try:
            (
                gamecount,
                bytes_per_game,
                positions_per_game,
                pieces_per_game,
                piecetypes_per_game,
            ) = self._import_estimates[:5]
        except TypeError as exc:
            raise DPTSizingError("No estimates available for sizing") from exc
        for item in self._import_estimates[:5]:
            if not isinstance(item, int):
                raise DPTSizingError("Value must be an 'int' instance")

        # Calculate number of standard profile games needed to generate
        # the number of index entries implied by the estimated profile
        # and number of games.
        d_count = (
            gamecount
            * (positions_per_game + pieces_per_game + piecetypes_per_game)
        ) // (
            POSITIONS_PER_GAME
            * (1 + PIECES_PER_POSITION + PIECES_TYPES_PER_POSITION)
        )

        # Calculate number of standard profile games needed to generate
        # the number of bytes implied by the estimated profile and number
        # of games.
        if bytes_per_game > BYTES_PER_GAME:
            b_count = int((gamecount * bytes_per_game) / BYTES_PER_GAME)
        else:
            b_count = gamecount

        # Use 'dict's because self._get_database_table_sizes() method
        # needs them internally, even though this case uses one file only.
        self._notional_record_counts = {
            GAMES_FILE_DEF: (b_count, d_count),
        }
        free = {}
        sizes, increments = self._get_database_table_sizes(
            files=self._notional_record_counts
        )

        append_text_only("")
        append_text("Standard profile game counts used in calculations.")
        append_text_only(
            " ".join(
                (
                    "Standard profile game count for data sizing:",
                    str(b_count),
                )
            )
        )
        append_text_only(
            " ".join(
                (
                    "Standard profile game count for index sizing:",
                    str(d_count),
                )
            )
        )
        append_text_only("")
        append_text_only(
            "".join(
                (
                    "A standard profile game is defined to have ",
                    str(POSITIONS_PER_GAME),
                    " positions, ",
                    str(PIECES_PER_POSITION),
                    " pieces per position, ",
                    str(PIECES_TYPES_PER_POSITION),
                    " piece types per position, and occupy ",
                    str(BYTES_PER_GAME),
                    " bytes.",
                )
            )
        )

        # Loops on sizes, increases, and free, dict objects removed because
        # this case does one file only.
        append_text_only("")
        append_text("Current file size and free space as pages.")
        bdsize = sizes[GAMES_FILE_DEF]
        bsize, bused, dsize, dused = bdsize
        bused = max(0, bused)
        free[GAMES_FILE_DEF] = (bsize - bused, dsize - dused)
        append_text_only(" ".join(("Current data area size:", str(bsize))))
        append_text_only(" ".join(("Current index area size:", str(dsize))))
        append_text_only(
            " ".join(("Current data area free:", str(bsize - bused)))
        )
        append_text_only(
            " ".join(("Current index area free:", str(dsize - dused)))
        )
        nr_count = self._notional_record_counts[GAMES_FILE_DEF]
        b_pages, d_pages = self.get_pages_for_record_counts(nr_count)
        append_text_only("")
        append_text("File space needed for import.")
        append_text_only(
            " ".join(("Estimated pages needed for data:", str(b_pages)))
        )
        append_text_only(
            " ".join(("Estimated pages needed for indexing:", str(d_pages)))
        )
        b_incr, d_incr = increments[GAMES_FILE_DEF]
        b_free, d_free = free[GAMES_FILE_DEF]

        # Save table B and D increases for import process to do later.
        # Save table B and D free for import process in case increase is not
        # enough and an adjustment has to be estimated if increase is 0.
        if increases is not None:
            increases[0] = b_incr
            increases[1] = d_incr
            increases[2] = b_free
            increases[3] = d_free

        append_text_only("")
        append_text("Planned file size increase and free space before import.")
        append_text_only(
            " ".join(("Planned increase in data pages:", str(b_incr)))
        )
        append_text_only(
            " ".join(("Planned increase in index pages:", str(d_incr)))
        )
        append_text_only(
            " ".join(("Free data pages before import:", str(b_incr + b_free)))
        )
        append_text_only(
            " ".join(("Free index pages before import:", str(d_incr + d_free)))
        )
        append_text_only("")
        append_text_only(
            "".join(
                (
                    "Comparison of the required and free data or index ",
                    "space may justify using the Increase Data and, or, ",
                    "Increase Index actions to get more space immediately ",
                    "given your knowledge of the PGN file being imported.",
                )
            )
        )

    # The attempt to generate a bz2 archive of the games database with
    # Python's builtin 'open' method fails because of a PermissionError.
    # The attempt to generate a DPT fast unload dump of the games
    # database fails because it is open in deferred update mode, and
    # attempting to open it temporarely in normal mode fails because the
    # audit file already exists (the database has to be closed down more
    # thoroughly than is convenient given need to keep other database
    # engines in play in the shared code).
    # Converting the archive to 'fast unload'  depend on being able to
    # produce a reliable crafted 'fast load' implementation to do the
    # import.  Till then the algorithm derived from 'text export' will
    # have to do.


class ChessDatabaseImportBackup(chessdptnofistat.ChessDatabase, ArchiveduDPT):
    """Access chess database to take backup before import."""

    # Set default parameters for fastload and fastunload use.
    def create_default_parms(self):
        """Create default parms.ini file for fast load/unload normal mode.

        This means transactions are disabled and a small number of buffers.

        """
        if not os.path.exists(self.parms):
            with open(self.parms, "w", encoding="iso-8859-1") as parms:
                parms.write("RCVOPT=X'00' " + os.linesep)
                parms.write("MAXBUF=100 " + os.linesep)


def chess_database_import_backup(
    dbpath, files, file=None, reporter=None, increases=None
):
    """Backup file before import to file."""
    budb = ChessDatabaseImportBackup(
        dbpath,
        allowcreate=True,
        sysfolder=os.path.join(dbpath, DPT_SYSFL_FOLDER),
    )
    budb.open_database(files=files)
    try:
        chess_du_backup_before_import(
            budb, file=file, reporter=reporter, increases=increases
        )
    finally:
        budb.close_database()


def chess_database_import(
    dbpath, files, *args, file=None, reporter=None, increases=None, **kwargs
):
    """Import to file."""
    cdb = ChessDatabase(dbpath, allowcreate=True)
    cdb.open_database(files=files)

    # Running out of table B pages gets a RuntimeError exception.
    # Running out of table D pages does not get an exception.
    # Both set the file as broken.
    try:
        chess_du_import(
            cdb,
            *args,
            file=file,
            reporter=reporter,
            increases=increases,
            **kwargs,
        )
    except RuntimeError as exc:
        if str(exc) != "File is full":
            raise

    # Necessary to update 'FISTAT' and 'FIFLAGS' information.
    cdb.close_database_contexts(files=files)
    cdb.open_database_contexts(files=files)

    try:
        for key in cdb.specification.keys():
            if key != file:
                continue
            parameters = cdb.table[key].get_file_parameters(cdb.dbenv)
            if parameters["FISTAT"][0] != FISTAT_DEFERRED_UPDATES:
                if reporter is not None:
                    reporter.append_text(
                        hex(FISTAT_DEFERRED_UPDATES).join(
                            (
                                "File broken during import (status not '",
                                "')",
                            )
                        )
                    )
                    reporter.append_text_only(parameters["FISTAT"][1])
                    reporter.append_text_only("")

            # Either table B or table D, but not both, may be marked full:
            # whichever happens first will cause the update to stop.
            # FISTAT bits can be turned off, but not on (1 to 0 is allowed).
            # FIFLAGS bits cannot be reset (1 to 0 or 0 to 1).
            if parameters["FIFLAGS"] & FIFLAGS_FULL_TABLEB:
                if reporter is not None:
                    reporter.append_text("File full: increase data size.")
                    reporter.append_text_only(
                        " ".join(
                            (
                                "Data size (too small) is",
                                str(parameters["BSIZE"]),
                                "pages.",
                            )
                        )
                    )
                if increases is None:
                    if reporter is not None:
                        reporter.append_text_only(
                            "Data size not changed: no increase specified."
                        )
                    reporter.append_text_only("")
                    break
                increment = increases[0] // 10
                if not increment:
                    increment = increases[2] // 10
                if not increment:
                    increment = increases[2]
                cdb.table[key].opencontext.Increase(increment, False)
                dinc = round(increment * cdb.specification[key][BTOD_FACTOR])
                cdb.table[key].opencontext.Increase(dinc, True)
                if reporter is not None:
                    viewer_resetter = cdb.dbenv.Core().GetViewerResetter()
                    reporter.append_text_only(
                        " ".join(
                            (
                                "Data size increased by",
                                str(increment),
                                "to",
                                viewer_resetter.View(
                                    "BSIZE", cdb.table[key].opencontext
                                ),
                                "pages.",
                            )
                        )
                    )
                    reporter.append_text_only(
                        " ".join(
                            (
                                "Index size increased by",
                                str(dinc),
                                "to",
                                viewer_resetter.View(
                                    "DSIZE", cdb.table[key].opencontext
                                ),
                                "pages to fit data size.",
                            )
                        )
                    )
                    reporter.append_text_only("File status is now:")
                    reporter.append_text_only(
                        viewer_resetter.View(
                            "FISTAT", cdb.table[key].opencontext
                        ),
                    )
                    reporter.append_text_only("")
            elif parameters["FIFLAGS"] & FIFLAGS_FULL_TABLED:
                if reporter is not None:
                    reporter.append_text("File full: increase index size.")
                    reporter.append_text_only(
                        " ".join(
                            (
                                "Index size (too small) is",
                                str(parameters["DSIZE"]),
                                "pages.",
                            )
                        )
                    )
                if increases is None:
                    if reporter is not None:
                        reporter.append_text_only(
                            "Index size not changed: no increase specified."
                        )
                    reporter.append_text_only("")
                    break
                viewer_resetter = cdb.dbenv.Core().GetViewerResetter()
                viewer_resetter.Reset(
                    "FISTAT",
                    str(parameters["FISTAT"][0] - FISTAT_PHYS_BROKEN),
                    cdb.table[key].opencontext,
                )
                increment = increases[1] // 10
                if not increment:
                    increment = increases[3] // 10
                if not increment:
                    increment = increases[3]
                cdb.table[key].opencontext.Increase(increment, True)
                if reporter is not None:
                    reporter.append_text_only(
                        " ".join(
                            (
                                "Index size increased by",
                                str(increment),
                                "to",
                                viewer_resetter.View(
                                    "DSIZE", cdb.table[key].opencontext
                                ),
                                "pages.",
                            )
                        )
                    )
                    reporter.append_text_only("File status is now:")
                    reporter.append_text_only(
                        viewer_resetter.View(
                            "FISTAT", cdb.table[key].opencontext
                        ),
                    )
                    reporter.append_text_only("")

            break

        else:
            if reporter is not None:
                reporter.append_text("File not open: status not checked.")
                reporter.append_text_only("")
    finally:
        cdb.close_database_contexts(files=files)


def chess_database_status_after_import(
    dbpath, files, file=None, reporter=None, increases=None, **kwargs
):
    """Recover games file from backup if file is broken."""
    budb = ChessDatabaseImportBackup(
        dbpath,
        allowcreate=True,
        sysfolder=os.path.join(dbpath, DPT_SYSFL_FOLDER),
    )
    if not os.path.exists(
        ".".join((os.path.join(dbpath, budb.import_backup_directory), "grd"))
    ):
        if reporter is not None:
            reporter.append_text("Backup '.grd' file does not exist.")
            reporter.append_text_only("Is backup, if it exists, trusted?")
            reporter.append_text_only("Backup, if it exists, is not deleted.")
            reporter.append_text_only("Database not recovered from backup.")
            reporter.append_text_only("")
        sys.exit(2)
    budb.open_database(files=files)
    try:
        games = budb.table[file]
        opencontext = games.opencontext
        viewer_resetter = budb.dbenv.Core().GetViewerResetter()
        if viewer_resetter.ViewAsInt("FISTAT", opencontext) == 0:  # Normal
            chess_du_delete_backup_after_import(
                budb,
                file=file,
                reporter=reporter,
                increases=increases,
                **kwargs,
            )
            return
        if reporter is not None:
            reporter.append_text("Initialize broken file.")
            reporter.append_text_only("")
        opencontext.Initialize()
        if reporter is not None:
            reporter.append_text("Recovering file from backup.")
            reporter.append_text_only("")

        # Does this step, working or not, belong in archivedudpt module?
        try:
            games.opencontext.Load(
                FLOAD_DEFAULT,
                0,
                None,
                os.path.join(
                    budb.home_directory, budb.import_backup_directory
                ),
            )
        except RuntimeError:
            fistat = viewer_resetter.ViewAsInt("FISTAT", games.opencontext)
            if (fistat & FIFLAGS_FULL_TABLEB) or (
                fistat & FIFLAGS_FULL_TABLED
            ):
                if reporter is not None:
                    reporter.append_text(
                        "File broken during recovery (status not '0x00')."
                    )
                    reporter.append_text_only("File status is now:")
                    reporter.append_text_only(
                        viewer_resetter.View(
                            "FISTAT", budb.table[file].opencontext
                        ),
                    )
                    reporter.append_text_only("")
                return
            raise

        if reporter is not None:
            reporter.append_text("File recovered.")
    finally:
        budb.close_database()


def chess_database_current_status(dbpath, files, file=None):
    """Exit process with non-zero code if games file is broken."""
    exit_code = 2
    budb = ChessDatabaseImportBackup(
        dbpath,
        allowcreate=True,
        sysfolder=os.path.join(dbpath, DPT_SYSFL_FOLDER),
    )
    budb.open_database(files=files)
    try:
        games = budb.table[file]
        opencontext = games.opencontext
        viewer_resetter = budb.dbenv.Core().GetViewerResetter()
        if viewer_resetter.ViewAsInt("FISTAT", opencontext) == 0:  # Normal
            exit_code = 0
        del opencontext
        budb.close_database_contexts(files=files)
    finally:
        budb.close_database()
    sys.exit(exit_code)



# chesslmdbdu.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for Symas LMMD."""

from solentware_base import lmdbdu_database

from ..shared.litedu import Litedu
from ..shared.alldu import chess_du, Alldu


class ChesslmdbduError(Exception):
    """Exception class for chesslmdbdu module."""


def chess_database_du(dbpath, *args, **kwargs):
    """Open database, import games and close database."""
    chess_du(ChessDatabase(dbpath, allowcreate=True), *args, **kwargs)


class ChessDatabase(Alldu, Litedu, lmdbdu_database.Database):
    """Provide custom deferred update for a database of games of chess."""

    def __init__(self, DBfile, **kargs):
        """Delegate with ChesslmdbduError as exception class."""
        super().__init__(DBfile, ChesslmdbduError, **kargs)

        # Assume DEFAULT_MAP_PAGES * 100 is enough for adding 64000 normal
        # sized games: the largest segment size holds 64000 games and a
        # commit is done after every segment.
        self._set_map_blocks_above_used_pages(100)

    def deferred_update_housekeeping(self):
        """Override to check map size and pages used expected page usage.

        The checks are done here because housekeeping happens when segments
        become full, a convenient point for commit and database resize.

        """
        self.commit()
        self._set_map_size_above_used_pages_between_transactions(100)
        self.start_transaction()



# blanktext.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Define tkinter.Text widget to be customized and used within ChessTab.

The positionrow.PositionRow and positionscore.PositionScore classes do not use
this class because their Text widgets come from solentware_grid.gui.DataRow.

"""

import tkinter
import enum

from solentware_bind.gui.bindings import Bindings

from .eventspec import EventSpec
from .displayitems import DisplayItemsStub
from .eventbinding import BlankTextEventBinding


# 'Tag' in these names refers to tags in Tk Text widgets, not PGN tags.
# NO_EDITABLE_TAGS and INITIAL_BINDINGS are used if _is_text_editable is False.
# All subclasses use DEFAULT_BINDINGS is _is_text_editable is True.
# Score uses NO_CURRENT_TOKEN, CURRENT_NO_TAGS, and SELECT_VARIATION, which
# other subclasses do not.
# GameEdit uses all the names.
class NonTagBind(enum.Enum):
    """Enumerate the tkinter.Text tag state.

    The tag state is used to control the event bindings for the widget.
    """

    # Current token is a move.
    NO_EDITABLE_TAGS = 1
    NO_CURRENT_TOKEN = 2
    DEFAULT_BINDINGS = 3
    INITIAL_BINDINGS = 4
    CURRENT_NO_TAGS = 5
    # Current token is a move with variations for next move, and the attempt
    # to go to next move was intercepted to choose which one.
    SELECT_VARIATION = 6


class BlankText(BlankTextEventBinding, Bindings):
    """Create Text widget with configuration shared by subclasses.

    The subclasses are cqltext.CQLText, querytext.QueryText, score.Score,
    and enginetext.EngineText.

    panel is used as the master argument for the tkinter Text and Scrollbar
    widgets created to display the statement text.

    items_manager is the ui attribute which tracks which EngineText instance is
    active (as defined by ui).

    Subclasses are responsible for providing a geometry manager.

    Attribute _is_text_editable is False to indicate text cannot be edited.

    Attribute _most_recent_bindings is set to indicate the initial set of
    event bindings.  Instances will override this as required.

    """

    # True means the content can be edited.
    _is_text_editable = False

    # Indicate the most recent set of bindings applied to score attribute.
    # Values are Tk tag names or members of NonTagBind enumeration.
    _most_recent_bindings = NonTagBind.INITIAL_BINDINGS

    def __init__(self, panel, items_manager=None, **ka):
        """Create widgets to display chess engine definition."""
        super().__init__(**ka)

        # May be worth using a Null() instance for these two attributes.
        if items_manager is None:
            items_manager = DisplayItemsStub()
        self.items = items_manager

        self.panel = panel
        self.score = tkinter.Text(
            master=self.panel,
            width=0,
            height=0,
            takefocus=tkinter.FALSE,
            undo=True,
            wrap=tkinter.WORD,
        )
        self.scrollbar = tkinter.Scrollbar(
            master=self.panel,
            orient=tkinter.VERTICAL,
            takefocus=tkinter.FALSE,
            command=self.score.yview,
        )
        self.score.configure(yscrollcommand=self.scrollbar.set)
        self.takefocus_widget = None

        # Keyboard actions do nothing by default.
        self._set_keypress_binding(switch=False)
        self.set_event_bindings_score(self._get_menubar_events())

        # The popup menus used by all subclasses.
        self.inactive_popup = None
        self.primary_activity_popup = None

    def _init_takefocus_widget(self, widget):
        """Initialize inactive_popup to popup if currently None."""
        # Allow callers to avoid the attribute-defined-outside-init message
        # if setting directly causes message to be given.
        self.takefocus_widget = widget

    def _init_inactive_popup(self, popup):
        """Initialize inactive_popup to popup if currently None."""
        # Allow callers to avoid the attribute-defined-outside-init message;
        # access-member-before-definition message is stopped too.
        # The assert ensures some __init__ has set inactive_popup to None,
        # although the real purpose is to insist it is not set already.
        assert hasattr(self, "inactive_popup") and self.inactive_popup is None
        self.inactive_popup = popup

    def set_event_bindings_score(self, bindings=(), switch=True):
        """Set bindings if switch is True or unset the bindings."""
        ste = self.try_event
        for sequence, function in bindings:
            self.bind(
                self.score,
                sequence[0],
                function=ste(function) if switch and function else "",
            )

    def _set_keypress_binding(self, function=None, bindings=(), switch=True):
        """Set bindings to function if switch is True or disable keypress."""
        if switch and function:
            stef = self.try_event(function)
            for sequence in bindings:
                self.bind(self.score, sequence[0], function=stef)
        else:
            stekb = self.try_event(self.press_break)
            for sequence in bindings:
                self.bind(self.score, sequence[0], function=stekb)

    def _get_menubar_events(self):
        """Return tuple of event binding definitions passed for menubar."""
        return ((EventSpec.score_enable_F10_menubar, self.press_none),)

    @staticmethod
    def press_break(event=None):
        """Do nothing and prevent event handling by next handlers."""
        del event
        return "break"

    @staticmethod
    def press_none(event=None):
        """Do nothing and allow event to be handled by next handler."""
        del event

    # This method arose when seeking clarity in the way popup menus were set,
    # and replaces lots of 'add_command' calls scattered all over.
    # Long term, either this method or _add_cascade_menu_to_popup will do all.
    def _set_popup_bindings(self, popup, bindings=(), index=tkinter.END):
        """Insert bindings in popup before index in popup."""
        # Default index is tkinter.END which seems to mean insert at end of
        # popup, not before last entry in popup as might be expected from the
        # way expressed in the 'Tk menu manual page' for index command.  (The
        # manual page describes 'end' in the context of 'none' for 'activate'
        # option.  It does make sense 'end' meaning after existing entries
        # when inserting entries.)

        for accelerator, function in bindings:
            popup.insert_command(
                index=index,
                label=accelerator[1],
                command=self.try_command(function, popup),
                accelerator=accelerator[2],
            )

    @staticmethod
    def give_focus_to_widget(event=None):
        """Do nothing and return 'break'.  Override in subclasses as needed."""
        del event
        return "break"

    @staticmethod
    def get_f10_popup_events(top_left, pointer):
        """Return tuple of event definitions to post popup menus.

        top_left is a method to post the menu at top left corner of widget.
        pointer is a method to post the menu at current pointer location.

        """
        return (
            (EventSpec.score_enable_F10_popupmenu_at_top_left, top_left),
            (EventSpec.score_enable_F10_popupmenu_at_pointer, pointer),
        )

    # Subclasses with database interfaces may override method.
    @staticmethod
    def _create_database_submenu(menu):
        """Do nothing.

        Subclasses should override this method as needed.

        """
        # pylint message assignment-from-none is reported when the return
        # value is assigned to a name.
        # The message is not reported if 'return None' is replaced like:
        # none = None
        # return none
        del menu

    def _post_menu(self, menu, create_menu, allowed=True, event=None):
        """Post the popup menu at current pointer location in widget.

        create_menu creates a popup menu if menu is None.
        event supplies the screen location for popup menu top left corner.
        allowed is True, default, permits display of menu.

        """
        del event
        if menu is None:
            menu = create_menu()
        if not allowed:
            return "break"
        menu.tk_popup(*self.score.winfo_pointerxy())

        # So 'Control-F10' does not fire 'F10' (menubar) binding too.
        return "break"

    @staticmethod
    def post_menu_at_top_left(menu, create_menu, allowed=True, event=None):
        """Post the popup menu at top left in widget.

        create_menu creates a popup menu if menu is None.
        event supplies the screen location for popup menu top left corner.
        allowed is True, default, permits display of menu.

        """
        if menu is None:
            menu = create_menu()
        if not allowed:
            return "break"
        menu.tk_popup(event.x_root - event.x, event.y_root - event.y)

        # So 'Shift-F10' does not fire 'F10' (menubar) binding too.
        return "break"

    def _is_active_item_mapped(self):
        """Return True if this widget is visible and is the active item."""
        if self.items.is_mapped_panel(self.panel):
            if self is not self.items.active_item:
                return False
        return True

    def _bind_for_primary_activity(self, switch=True):
        """Set (switch True) or clear bindings for main actions when active.

        If bool(switch) is true, clear the most recently set bindings first.

        """
        if switch:
            self.token_bind_method[self._most_recent_bindings](self, False)
            self._most_recent_bindings = NonTagBind.NO_EDITABLE_TAGS
        self._bind_for_set_primary_activity_bindings(switch)

    def bind_for_initial_state(self, switch=True):
        """Clear the most recently set bindings if bool(switch) is True.

        Assume not setting new bindings leaves widget in initial state.

        If bool(switch) is False, nothing is done.

        """
        if switch:
            self.token_bind_method[self._most_recent_bindings](self, False)
            self._most_recent_bindings = NonTagBind.INITIAL_BINDINGS

    # Dispatch dictionary for token binding selection.
    # Keys are the possible values of self._most_recent_bindings.
    token_bind_method = {
        NonTagBind.NO_EDITABLE_TAGS: _bind_for_primary_activity,
        NonTagBind.INITIAL_BINDINGS: bind_for_initial_state,
    }

    def _create_primary_activity_popup(self):
        """Create and return popup menu with primary and database commands.

        The primary commands are set by subclasses and should be related
        to their main function.

        By default no database commands are set.  Subclasses should
        override the create_database_submanu method as needed.

        """
        # pylint message access-member-before-definition.
        # Initialize primary_activity_popup moved to blanktext from
        # enginetext module.
        assert self.primary_activity_popup is None
        popup = tkinter.Menu(master=self.score, tearoff=False)
        self._set_popup_bindings_get_primary_activity_events(popup)
        # pylint message assignment-from-none is false positive.
        # However it is sensible to do an isinstance test.
        database_submenu = self._create_database_submenu(popup)
        if isinstance(database_submenu, tkinter.Menu):
            popup.add_cascade(label="Database", menu=database_submenu)
        self.primary_activity_popup = popup
        return popup



# chessscore.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Demonstrate chess score class and methods to display PGN text."""
# This module replaces one which used an obsolete version of Game class,
# probably from around 2008.

import tkinter


class DuckBoard:
    """Helper to allow Score instances to work without a board.

    ui is the real ui to which Score methods refer as necessary.

    """

    boardsquares = {}

    def __init__(self, ui):
        """Set reference to ui User Interface object."""
        self.ui = ui

    def set_board(self, board):
        """Do nothing because board display is not implemented."""


class DuckUI:
    """Helper to allow Score instances to work without a real UI."""

    def __init__(self, panel, *a, **k):
        """Note panel as reference point to get top level widget."""
        self.top_pw = panel

    def get_toplevel(self):
        """Return the top level widget."""
        return self.top_pw.winfo_toplevel()

    def get_export_filename_for_single_item(self, *a, **k):
        """Display dialogue saying export not supported here."""
        tkinter.messagebox.showinfo(
            parent=self.get_toplevel(),
            title="Export Game",
            message="Export not supported in this UI",
        )


if __name__ == "__main__":
    from pgn_read.core.parser import PGN

    from ..gui import fonts
    from ..gui.score import Score

    root = tkinter.Tk()
    root.wm_minsize(width=900, height=300)
    f = fonts.make_chess_fonts(root)
    root.wm_title("Demonstrate Score without Board")
    root.pack_propagate(False)

    # Use DuckBoard.
    s = Score(root, DuckBoard(DuckUI(root)))

    del f
    s.collected_game = next(
        PGN().read_games(
            "".join(
                (
                    '[Event"National Club: Gosport - Wood Green"]',
                    '[Site"Gosport"]',
                    '[Date"1989-05-07"]',
                    '[Round"QFinal"]',
                    '[White"Sowray P J"]',
                    '[Black"Marsh R"]',
                    '[Result"1-0"]',
                    "e4(d4d5c4e6Nc3)c6d4d5exd5cxd5c4Nf6c5e6Nc3b6b4a5Bf4",
                    "axb4Nb5Na6Qa4;comment to eol\nBd7",
                    "Bc7Nxc5Qd1Qc8dxc5bxc5Nf3Qb7Nd6Bxd6Bxd6Qb6",
                    "Be5Ke7Be2Ne4O-Of6Bb2Nc3",
                    "Bxc3bxc3Qd3(Qb3)Ra3Rfb1Qa7Qc2g6Rb3d4Bc4",
                    "Rxb3Bxb3Qa6a4Rb8a5e5Bd5",
                    "\n%The escape sequence\nRb2",
                    "Qe4Bf5Qh4Qd3(c2(g5)Nd2Qxa5Rxa5Rb1",
                    "{Comment\ncontaining newline}Nf1",
                    "(Nxb1c1=Q)Rxf1Kxf1Bd3)g4Rb1",
                    "Rxb1Qxb1Kg2Kd6Qxf6Kxd5<reserved\n\nfor future use>Qxe5",
                    "Kc6gxf5Qxf5Qe8Kc7Qe7Kc8Ne5c2Qxc5Kd8Qxd4",
                    "Ke8Qe3Kf8Kg3Qc8Nd3Kg8f4Qc6Nc1Qa4Qb3",
                    "1-0",
                    "\n",
                )
            )
        )
    )

    # This works, but set_and_tag_item_text is the method used normally.
    # The tkinter.Text widget must get configured 'state="disabled"' somewhere
    # because get_keypress_suppression_events is not called anywhere in
    # chesstab.  (Ah! set_and_tag_item_text).
    s.map_game()
    s.set_event_bindings_score(s.get_keypress_suppression_events())
    s.bind_for_move()
    # s.set_and_tag_item_text()

    # s.score is not in any container, except root, so use directly for focus
    # and geometry.
    # The get_top_widget method and takefocus_widget attribute are defined in
    # Game class to keep them out of AnalysisScore instances.
    s.score.pack(fill=tkinter.BOTH, expand=tkinter.TRUE)
    s.score.pack_propagate(False)
    s.score.focus_set()
    del s

    root.mainloop()



# dbtkinterdu_file_fpd.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN with db_tkinter.chessdbtkinterdu to one file per database."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..db_tkinter.chessdbtkinterdu import ChessDatabase

    FileWidget(ChessDatabase, "db_tcl", file_per_database=True)



# chesslmdb.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database using Symas LMMD."""

from solentware_base import lmdb_database

from ..core.filespec import FileSpec
from ..basecore import database


class ChessDatabase(database.Database, lmdb_database.Database):
    """Provide access to a database of games of chess."""

    _deferred_update_process = "chesstab.lmdb.chesslmdbdu"

    def __init__(
        self,
        DBfile,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        Arguments are passed through to superclass __init__.

        """
        dbnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )
        super().__init__(dbnames, folder=DBfile, **kargs)

        # Allow space for lots of chess engine analysis.
        self._set_map_blocks_above_used_pages(200)

    def _delete_database_names(self):
        """Override and return tuple of filenames to delete."""
        return (self.database_file,)

    def checkpoint_before_close_dbenv(self):
        """Override.  Hijack method to set map size to file size.

        Reverse, to the extent possible, the increase in map size done
        when the database was opened.

        """
        self._set_map_size_above_used_pages_between_transactions(0)



# repertoiredisplay.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widgets to display and edit repertoires.

These four classes display PGN text for games in the main window: they are
used in the gamelistgrid module.

The _RepertoireDisplay class provides attributes and behaviour shared by the
RepertoireDisplay, RepertoireDisplayInsert, and RepertoireDisplayEdit,
classes.  It also provides properties to support implementation of behaviour
shared with the CQL*, Game*, and Query*, classes.

The RepertoireDisplay, RepertoireDisplayInsert, and RepertoireDisplayEdit,
classes are subclasses of the relevant ShowPGN, InsertPGN, EditPGN, and
DisplayPGN, classes from the displaypgn module; to implement behaviour shared
with all text widgets in the main display (that includes widgets displaying
text).

"""

from solentware_grid.core.dataclient import DataNotify

from solentware_bind.gui.bindings import Bindings

from ..core.constants import TAG_OPENING
from .repertoire import Repertoire
from .repertoireedit import RepertoireEdit
from ..core.chessrecord import ChessDBrecordRepertoireUpdate
from .eventspec import EventSpec
from .display import Display
from .displaypgn import ShowPGN, InsertPGN, EditPGN, DisplayPGN
from .game import Game


class _RepertoireDisplay(ShowPGN, Game, Bindings, DataNotify, Display):
    """Extend and link PGN repertoire text to database.

    sourceobject - link to database.

    Attribute binding_labels specifies the order navigation bindings appear
    in popup menus.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Attribute pgn_score_tags provides the PGN tag names used in widget titles
    and message text.  It is the opening PGN tag defined in ChessTab.

    Attribute pgn_score_source provides the error key value to index a
    PGN game score with errors.

    Attribute pgn_score_updater provides the class used to process PGN text
    into a database update.

    """

    binding_labels = (
        EventSpec.navigate_to_position_grid,
        EventSpec.navigate_to_active_game,
        EventSpec.navigate_to_game_grid,
        EventSpec.navigate_to_repertoire_grid,
        EventSpec.repertoiredisplay_to_previous_repertoire,
        EventSpec.analysis_to_scoresheet,
        EventSpec.repertoiredisplay_to_next_repertoire,
        EventSpec.navigate_to_repertoire_game_grid,
        EventSpec.scoresheet_to_analysis,
        EventSpec.navigate_to_partial_grid,
        EventSpec.navigate_to_active_partial,
        EventSpec.navigate_to_partial_game_grid,
        EventSpec.navigate_to_selection_rule_grid,
        EventSpec.navigate_to_active_selection_rule,
        EventSpec.tab_traverse_backward,
        EventSpec.tab_traverse_forward,
    )

    # These exist so the insert_game_database methods in _RepertoireDisplay and
    # gamedisplay._GameDisplay, and delete_game_database in RepertoireDisplay
    # and gameedisplay.GameDisplay, can be modified and replaced by single
    # copies in the displaypgn.ShowPGN class.
    # See mark_partial_positions_to_be_recalculated() method too.
    # The names need to be more generic to make sense in cql, engine, and
    # query, context.
    pgn_score_name = "repertoire"
    pgn_score_source = "No opening name"
    pgn_score_tags = (TAG_OPENING,)
    pgn_score_updater = ChessDBrecordRepertoireUpdate

    def __init__(self, sourceobject=None, **ka):
        """Extend and link repertoire to database."""
        super().__init__(**ka)
        self.blockchange = False
        if self.ui.base_repertoires.datasource:
            self.set_data_source(self.ui.base_repertoires.get_data_source())
        self.sourceobject = sourceobject

    # Could be put in game.Repertoire class to hide the game.Game version.
    # Here can be justified because purpose is allow some methods to be moved
    # to displaypgn.ShowPGN class.
    @property
    def ui_displayed_items(self):
        """Return manager of widgets displaying a repertoire record."""
        return self.ui.repertoire_items

    # Defined so _cycle_item and give_focus_to_widget methods can be shared by
    # gamedisplay._GameDisplay and repertoiredisplay._RepertoireDisplay
    # classes.
    @property
    def ui_configure_item_list_grid(self):
        """Return function to configure repertoire grid to fit text."""
        return self.ui.configure_repertoire_grid

    # ui_base_table and mark_partial_positions_to_be_recalculated defined
    # so insert_game_database method can be shared by gamedisplay._GameDisplay
    # and repertoiredisplay._RepertoireDisplay classes.
    # See class attributes pgn_score_name and pgn_score_source too.

    @property
    def ui_base_table(self):
        """Return the User Interface TagRosterGrid object."""
        return self.ui.base_repertoires

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.games_and_repertoires_in_toplevels

    def mark_partial_positions_to_be_recalculated(self, datasource=None):
        """Do nothing.  Exists for compatibility with game display."""

    def _get_navigation_events(self):
        """Return event description tuple for navigation from repertoire."""
        return (
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (EventSpec.navigate_to_active_game, self.set_focus_gamepanel_item),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item,
            ),
            (
                EventSpec.repertoiredisplay_to_previous_repertoire,
                self._prior_item,
            ),
            (EventSpec.repertoiredisplay_to_next_repertoire, self._next_item),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            # No traverse_round because Alt-F8 toggles repertoire and analysis.
        )

    def delete_item_view(self, event=None):
        """Remove repertoire item from screen."""
        del event
        self.set_data_source()
        self.ui.delete_repertoire_view(self)

    def on_game_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if self.sourceobject is not None:
            self._patch_pgn_score_to_fit_record_change_and_refresh_grid(
                self.ui.repertoire_games, instance
            )

    def on_repertoire_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if self.sourceobject is not None:
            if (
                instance.key == self.sourceobject.key
                and self.datasource.dbname == self.sourceobject.dbname
                and self.datasource.dbset == self.sourceobject.dbset
            ):
                self.blockchange = True

    def generate_popup_navigation_maps(self):
        """Return genenal and current widget navigation binding maps."""
        navigation_map = dict(self._get_navigation_events())
        local_map = {
            EventSpec.scoresheet_to_analysis: self.analysis_current_item,
        }
        return navigation_map, local_map


class RepertoireDisplay(
    _RepertoireDisplay, DisplayPGN, ShowPGN, Repertoire, DataNotify
):
    """Display a repertoire from a database allowing delete and insert."""

    # Allow for structure difference between RepertoireDisplay and GameDisplay
    # versions of delete_game_database.
    # Method comments suggest a problem exists which needs fixing.
    # Existence of this method prevents delete_game_database being used by
    # instances of superclasses of RepertoireDisplay, emulating the behaviour
    # before introduction of displaypgn module.
    @staticmethod
    def pgn_score_original_value(original_value):
        """Set source name for original_value object."""
        # currently attracts "AttributeError: 'ChessDBvalueGameTags' has
        # no attribute 'gamesource'.
        # original.value.set_game_source(self.sourceobject.value.gamesource)
        original_value.set_game_source("No opening name")

    def _create_primary_activity_popup(self):
        """Delegate then add close command to popup and return popup menu."""
        popup = super()._create_primary_activity_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_select_move_popup(self):
        """Delegate then add close command to popup and return popup menu."""
        popup = super()._create_select_move_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup


class RepertoireDisplayInsert(
    _RepertoireDisplay, InsertPGN, ShowPGN, RepertoireEdit, DataNotify
):
    """Display a repertoire from a database allowing insert.

    RepertoireEdit provides the widget and _RepertoireDisplay the database
    interface.

    """


class RepertoireDisplayEdit(EditPGN, RepertoireDisplayInsert):
    """Display a repertoire from a database allowing edit and insert."""

    # Allow for structure difference between RepertoireDisplay and GameDisplay
    # versions of delete_game_database.
    # Method comments suggest a problem exists which needs fixing.
    # Existence of this method prevents delete_game_database being used by
    # instances of superclasses of RepertoireDisplay, emulating the behaviour
    # before introduction of displaypgn module.
    @staticmethod
    def pgn_score_original_value(original_value):
        """Set source name for original_value object."""
        # currently attracts "AttributeError: 'ChessDBvalueGameTags' has
        # no attribute 'gamesource'.
        # original.value.set_game_source(self.sourceobject.value.gamesource)
        original_value.set_game_source("No opening name")

    # _set_properties_on_grids defined so update_game_database method can be
    # shared by repertoiredisplay.RepertoireDisplayEdit and
    # gamedisplay.GameDisplayEdit classes.
    # See class attributes pgn_score_name and pgn_score_source too.
    # The property which returns self.ui.base_repertoires is ignored because
    # the GameDisplayEdit version of the method sets properties on all grids.
    def set_properties_on_game_grids(self, newkey):
        """Set grid row properties of row for newkey record."""
        self.ui.base_repertoires.set_properties(newkey)
